import json
import os.path
import sys
from importlib.metadata import distribution

VERSION = (1, 14, 0)
__version__ = ".".join(map(str, VERSION))


PKG_ROOT = os.path.dirname(__file__)
DATA_DIR = os.path.join(PKG_ROOT, "data-files")


def is_installed():
    if not is_dev():
        return True
    if _is_venv():
        return True
    return _is_devel_install()


def is_dev():
    parent = os.path.dirname(PKG_ROOT)
    return os.path.exists(os.path.join(parent, "pyproject.toml"))


def _is_venv():
    if sys.base_prefix == sys.prefix:
        return False
    return True


def _is_devel_install():
    # pip install -e <path-to-git-checkout> will do a "devel" install.
    # This means it creates a link back to the checkout instead
    # of copying the files.

    direct_url = distribution("pyperformance").read_text("direct_url.json")
    if direct_url:
        return json.loads(direct_url).get("dir_info", {}).get("editable", False)
    return False




"""Test the performance of simple HTTP serving and client using the Tornado
framework.

A trivial "application" is generated which generates a number of chunks of
data as a HTTP response's body.
"""

import sys
import socket

import pyperf

from tornado.httpclient import AsyncHTTPClient
from tornado.httpserver import HTTPServer
from tornado.gen import coroutine
from tornado.ioloop import IOLoop
from tornado.netutil import bind_sockets
from tornado.web import RequestHandler, Application


HOST = "127.0.0.1"
FAMILY = socket.AF_INET

CHUNK = b"Hello world\n" * 1000
NCHUNKS = 5

CONCURRENCY = 150


class MainHandler(RequestHandler):

    @coroutine
    def get(self):
        for i in range(NCHUNKS):
            self.write(CHUNK)
            yield self.flush()

    def compute_etag(self):
        # Overriden to avoid stressing hashlib in this benchmark
        return None


def make_application():
    return Application([
        (r"/", MainHandler),
    ])


def make_http_server(request_handler):
    server = HTTPServer(request_handler)
    sockets = bind_sockets(0, HOST, family=FAMILY)
    assert len(sockets) == 1
    server.add_sockets(sockets)
    sock = sockets[0]
    return server, sock


def bench_tornado(loops):
    server, sock = make_http_server(make_application())
    host, port = sock.getsockname()
    url = "http://%s:%s/" % (host, port)
    namespace = {}

    @coroutine
    def run_client():
        client = AsyncHTTPClient()
        range_it = range(loops)
        t0 = pyperf.perf_counter()

        for _ in range_it:
            futures = [client.fetch(url) for j in range(CONCURRENCY)]
            for fut in futures:
                resp = yield fut
                buf = resp.buffer
                buf.seek(0, 2)
                assert buf.tell() == len(CHUNK) * NCHUNKS

        namespace['dt'] = pyperf.perf_counter() - t0
        client.close()

    IOLoop.current().run_sync(run_client)
    server.stop()

    return namespace['dt']


if __name__ == "__main__":
    # 3.8 changed the default event loop to ProactorEventLoop which doesn't
    # implement everything required by tornado and breaks this benchmark.
    # Restore the old WindowsSelectorEventLoop default for now.
    # https://bugs.python.org/issue37373
    # https://github.com/python/pyperformance/issues/61
    # https://github.com/tornadoweb/tornado/pull/2686
    if sys.platform == 'win32' and sys.version_info[:2] >= (3, 8):
        import asyncio
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

    runner = pyperf.Runner()
    runner.metadata['description'] = ("Test the performance of HTTP requests "
                                      "with Tornado.")
    runner.bench_time_func('tornado_http', bench_tornado)



import errno
import importlib.util
import os
import os.path
import shlex
import shutil
import subprocess
import sys
import tempfile
import unittest

TESTS_ROOT = os.path.realpath(os.path.dirname(__file__))
DATA_DIR = os.path.join(TESTS_ROOT, "data")
REPO_ROOT = os.path.dirname(os.path.dirname(TESTS_ROOT))
DEV_SCRIPT = os.path.join(REPO_ROOT, "dev.py")


def run_cmd(cmd, *args, capture=None, onfail="exit", verbose=True):
    # XXX Optionally write the output to a file.
    argv = (cmd,) + args
    if not all(a and isinstance(a, str) for a in argv):
        raise TypeError(f"all args must be non-empty strings, got {argv}")
    argv_str = " ".join(shlex.quote(a) for a in argv)

    kwargs = dict(
        cwd=REPO_ROOT,
    )
    if capture is True:
        capture = "both"
    if capture in ("both", "combined", "stdout"):
        kwargs["stdout"] = subprocess.PIPE
    if capture in ("both", "stderr"):
        kwargs["stderr"] = subprocess.PIPE
    elif capture == "combined":
        kwargs["stderr"] = subprocess.STDOUT
    if capture:
        kwargs["encoding"] = "utf-8"

    if verbose:
        print(f"(tests) Execute: {argv_str}", flush=True)
    proc = subprocess.run(argv, **kwargs)

    exitcode = proc.returncode
    if exitcode:
        if onfail == "exit":
            sys.exit(exitcode)
        elif onfail == "raise":
            raise Exception(f'"{argv_str}" failed (exitcode {exitcode})')
        elif onfail is not None:
            raise NotImplementedError(repr(onfail))
    if "stdout" not in kwargs or "stderr" not in kwargs:
        print("", flush=True)

    return exitcode, proc.stdout, proc.stderr


def _resolve_venv_python(venv):
    if os.name == "nt":
        basename = os.path.basename(sys.executable)
        venv_python = os.path.join(venv, "Scripts", basename)
    else:
        venv_python = os.path.join(venv, "bin", "python3")
    return venv_python


def create_venv(root=None, python=sys.executable, *, verbose=False):
    if not root:
        tmpdir = tempfile.mkdtemp()
        root = os.path.join(tmpdir, "venv")

        def cleanup():
            return shutil.rmtree(tmpdir)
    else:

        def cleanup():
            return None

    run_cmd(
        python or sys.executable,
        "-m",
        "venv",
        root,
        capture=not verbose,
        onfail="raise",
        verbose=verbose,
    )
    return root, _resolve_venv_python(root), cleanup


class CleanupFile:
    def __init__(self, filename):
        self.filename = filename

    def __repr__(self):
        return f"{type(self).__name__}({self.filename!r})"

    def __str__(self):
        return self.filename

    def __enter__(self):
        return self.filename

    def __exit__(self, *args):
        self.cleanup()

    def cleanup(self):
        try:
            os.unlink(self.filename)
        except OSError as exc:
            if exc.errno != errno.ENOENT:
                raise


#############################
# testing fixtures, mixins, and helpers


def apply_to_test_methods(cls, decorator):
    for name, func in vars(cls).items():
        if not name.startswith("test_"):
            continue
        func = decorator(func)
        setattr(cls, name, func)


def mark(label, func=None):
    """Mark the function/class with the given label.

    This may be used as a decorator.
    """
    if func is None:

        def decorator(func):
            return mark(label, func)

        return decorator
    if isinstance(func, type):
        cls = func
        apply_to_test_methods(cls, mark(label))
        return cls
    try:
        func._pyperformance_test_labels.append(label)
    except AttributeError:
        func._pyperformance_test_labels = [label]
    return func


#############################
# functional tests

CPYTHON_ONLY = unittest.skipIf(
    sys.implementation.name != "cpython",
    "CPython-only",
)
NON_WINDOWS_ONLY = unittest.skipIf(os.name == "nt", "skipping Windows")


# XXX Provide a way to run slow tests.
def SLOW(f):
    return unittest.skip("way too slow")(mark("slow", f))


class Functional:
    """A mixin for functional tests.

    In this context, "functional" means the test touches the filesystem,
    uses the network, creates subprocesses, etc.  This contrasts with
    true "unit" tests, which are constrained strictly to code execution.
    """

    @classmethod
    def resolve_tmp(cls, *relpath, unique=False):
        try:
            tmpdir = cls._tmpdir
        except AttributeError:
            tmpdir = cls._tmpdir = tempfile.mkdtemp()
            cls.addClassCleanup(lambda: shutil.rmtree(tmpdir))
        if unique:
            tmpdir = tempfile.mkdtemp(dir=tmpdir)
        return os.path.join(tmpdir, *relpath)

    @classmethod
    def run_python(cls, *args, require_venv=True, **kwargs):
        python = getattr(cls, "_venv_python", None)
        if not python:
            if require_venv:
                raise Exception("cls.ensure_venv() must be called first")
            python = sys.executable
        # We always use unbuffered stdout/stderr to simplify capture.
        return run_cmd(python, "-u", *args, **kwargs)

    @classmethod
    def run_module(cls, module, *args, **kwargs):
        return cls.run_python("-m", module, *args, **kwargs)

    @classmethod
    def run_pip(cls, *args, **kwargs):
        return cls.run_module("pip", *args, **kwargs)

    @classmethod
    def ensure_venv(cls):
        if getattr(cls, "_tests_venv", None):
            # ensure_venv() already ran.
            return

        if sys.prefix != sys.exec_prefix:
            # We are already running in a venv and assume it is ready.
            cls._tests_venv = sys.prefix
            cls._venv_python = sys.executable
            return

        cls._tests_venv = cls._resolve_tests_venv()
        venv_python = _resolve_venv_python(cls._tests_venv)
        if os.path.exists(venv_python):
            cls._venv_python = venv_python
            return

        print("#" * 40)
        print("# creating a venv")
        print("#" * 40)
        print()

        # Create the venv and update it.
        # XXX Ignore the output (and optionally log it).
        create_venv(cls._tests_venv, verbose=True)
        cls._venv_python = venv_python
        cls.run_pip("install", "--upgrade", "pip")
        cls.run_pip("install", "--upgrade", "setuptools", "wheel")

        print("#" * 40)
        print("# DONE: creating a venv")
        print("#" * 40)
        print()

    @classmethod
    def _resolve_tests_venv(cls):
        root = os.environ.get("PYPERFORMANCE_TESTS_VENV")
        if not root:
            root = "<reuse>"
        elif not root.startswith("<") and not root.endswith(">"):
            # The user provided an actual root dir.
            return root

        if root == "<temp>":
            # The user is forcing a temporary venv.
            return cls.resolve_tmp("venv")
        elif root == "<fresh>":
            # The user is forcing a fresh venv.
            fresh = True
        elif root == "<reuse>":
            # This is the default.
            fresh = False
        else:
            raise NotImplementedError(repr(root))

        # Resolve the venv root to re-use.
        spec = importlib.util.spec_from_file_location("dev", DEV_SCRIPT)
        dev = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(dev)
        root = dev.resolve_venv_root("tests")

        if fresh:
            if os.path.exists(root):
                print("(refreshing existing venv)")
                print()
            try:
                shutil.rmtree(root)
            except FileNotFoundError:
                pass
        return root



import os
import os.path
import sys
import types
import unittest

import pyperformance._venv
from pyperformance import tests


def new_venv_config(
    root,
    home=None,
    version=None,
    prompt=None,
    system_site_packages=False,
    executable=None,
    command=None,
    *,
    old=sys.version_info < (3, 11),
):
    if not home:
        home = os.path.dirname(sys.executable)
    if not version:
        version = ".".join(str(v) for v in sys.version_info[:3])
    if not isinstance(system_site_packages, str):
        system_site_packages = "true" if system_site_packages else "false"
    system_site_packages = system_site_packages == "true"
    if not old:
        if executable is None:
            executable = sys.executable
        if command is None:
            command = f"{executable} -m venv {root}"
    return types.SimpleNamespace(
        home=home,
        version=version,
        system_site_packages=system_site_packages,
        prompt=prompt,
        executable=executable,
        command=command,
    )


def render_venv_config(cfg):
    lines = [
        f"home = {cfg.home}",
        f"version = {cfg.version}",
        f"include-system-site-packages = {cfg.system_site_packages}",
    ]
    if cfg.prompt is not None:
        lines.append(f"prompt = {cfg.prompt}")
    if cfg.executable is not None:
        lines.append(f"executable = {cfg.executable}")
    if cfg.command is not None:
        lines.append(f"command = {cfg.command}")
    return "".join(line + os.linesep for line in lines)


class VenvConfigTests(tests.Functional, unittest.TestCase):
    # Note that we do not call self.ensure_venv().

    def generate_config(self, root, **kwargs):
        if not os.path.isabs(root):
            root = self.resolve_tmp(root)
        cfg = new_venv_config(root, **kwargs)
        text = render_venv_config(cfg)

        os.makedirs(root)
        filename = os.path.join(root, "pyvenv.cfg")
        with open(filename, "w", encoding="utf-8") as outfile:
            outfile.write(text)
        return cfg, filename, root

    def test_read(self):
        expected, _, root = self.generate_config("spam")

        cfg = pyperformance._venv.read_venv_config(root)

        self.assertEqual(vars(cfg), vars(expected))

    def test_parse(self):
        expected = new_venv_config("spam")
        text = render_venv_config(expected)

        cfg = pyperformance._venv.parse_venv_config(text)

        self.assertEqual(vars(cfg), vars(expected))




"""Benchmark how quickly Python's regex implementation can compile regexes.

We bring in all the regexes used by the other regex benchmarks, capture them by
stubbing out the re module, then compile those regexes repeatedly. We muck with
the re module's caching to force it to recompile every regex we give it.
"""

# Python imports
import re

# Local imports
import pyperf


def capture_regexes():
    regexes = []

    real_compile = re.compile
    real_search = re.search
    real_sub = re.sub

    def capture_compile(regex, flags=0):
        regexes.append((regex, flags))
        return real_compile(regex, flags)

    def capture_search(regex, target, flags=0):
        regexes.append((regex, flags))
        return real_search(regex, target, flags)

    def capture_sub(regex, *args):
        regexes.append((regex, 0))
        return real_sub(regex, *args)

    re.compile = capture_compile
    re.search = capture_search
    re.sub = capture_sub
    try:
        import bm_regex_effbot
        bm_regex_effbot.bench_regex_effbot(1)

        import bm_regex_v8
        bm_regex_v8.bench_regex_v8(1)
    finally:
        re.compile = real_compile
        re.search = real_search
        re.sub = real_sub
    return regexes


def bench_regex_compile(loops, regexes):
    range_it = range(loops)
    t0 = pyperf.perf_counter()

    for _ in range_it:
        for regex, flags in regexes:
            re.purge()
            # ignore result (compiled regex)
            re.compile(regex, flags)

    return pyperf.perf_counter() - t0


if __name__ == "__main__":
    runner = pyperf.Runner()
    runner.metadata['description'] = "Test regex compilation performance"

    regexes = capture_regexes()
    runner.bench_time_func('regex_compile', bench_regex_compile, regexes)



"""Benchmark for the base64 module's primary public APIs.

Tests encoding and decoding performance across various variants
and data sizes, split into _small (balanced small data) and _large variants.

Small weighs towards measuring overhead, large measures the core algorithm
loop implementation.
"""

import base64
import random
import pyperf


# Generate test data with a fixed seed for reproducibility
random.seed(12345)
DATA_TINY = random.randbytes(20)
DATA_SMALL = random.randbytes(127)  # odd on purpose
DATA_MEDIUM = random.randbytes(3072)
DATA_9K = random.randbytes(9000)
DATA_LARGE = random.randbytes(102400)
DATA_HUGE = random.randbytes(1048574)  # 1M-2

# Pre-encoded data for decode benchmarks
B64_TINY = base64.b64encode(DATA_TINY)
B64_SMALL = base64.b64encode(DATA_SMALL)
B64_MEDIUM = base64.b64encode(DATA_MEDIUM)
B64_9K_STR = base64.b64encode(DATA_9K).decode('ascii')
B64_LARGE = base64.b64encode(DATA_LARGE)
B64_HUGE = base64.b64encode(DATA_HUGE)

B64_URLSAFE_TINY = base64.urlsafe_b64encode(DATA_TINY)
B64_URLSAFE_SMALL = base64.urlsafe_b64encode(DATA_SMALL)
B64_URLSAFE_MEDIUM = base64.urlsafe_b64encode(DATA_MEDIUM)
B64_URLSAFE_9K_STR = base64.urlsafe_b64encode(DATA_9K).decode('ascii')

B32_TINY = base64.b32encode(DATA_TINY)
B32_SMALL = base64.b32encode(DATA_SMALL)
B32_MEDIUM = base64.b32encode(DATA_MEDIUM)
B32_9K_STR = base64.b32encode(DATA_9K).decode('ascii')
B32_LARGE = base64.b32encode(DATA_LARGE)
B32_HUGE = base64.b32encode(DATA_HUGE)

B16_TINY = base64.b16encode(DATA_TINY)
B16_SMALL = base64.b16encode(DATA_SMALL)
B16_MEDIUM = base64.b16encode(DATA_MEDIUM)
B16_9K_STR = base64.b16encode(DATA_9K).decode('ascii')
B16_LARGE = base64.b16encode(DATA_LARGE)
B16_HUGE = base64.b16encode(DATA_HUGE)

A85_TINY = base64.a85encode(DATA_TINY)
A85_SMALL = base64.a85encode(DATA_SMALL)
A85_MEDIUM = base64.a85encode(DATA_MEDIUM)
A85_9K_STR = base64.a85encode(DATA_9K).decode('ascii')
A85_LARGE = base64.a85encode(DATA_LARGE)
A85_HUGE = base64.a85encode(DATA_HUGE)

B85_TINY = base64.b85encode(DATA_TINY)
B85_SMALL = base64.b85encode(DATA_SMALL)
B85_MEDIUM = base64.b85encode(DATA_MEDIUM)
B85_9K_STR = base64.b85encode(DATA_9K).decode('ascii')
B85_LARGE = base64.b85encode(DATA_LARGE)
B85_HUGE = base64.b85encode(DATA_HUGE)


# --- Base64 ---

def bench_b64_small(loops):
    range_it = range(loops)
    t0 = pyperf.perf_counter()
    for _ in range_it:
        for _ in range(450):
            base64.b64encode(DATA_TINY)
            base64.b64decode(B64_TINY)
        for _ in range(71):
            base64.b64encode(DATA_SMALL)
            base64.b64decode(B64_SMALL)
        for _ in range(3):
            base64.b64encode(DATA_MEDIUM)
            base64.b64decode(B64_MEDIUM)
        base64.b64encode(DATA_9K)
        base64.b64decode(B64_9K_STR)
    return pyperf.perf_counter() - t0


def bench_b64_large(loops):
    range_it = range(loops)
    t0 = pyperf.perf_counter()
    for _ in range_it:
        for _ in range(10):
            base64.b64encode(DATA_LARGE)
            base64.b64decode(B64_LARGE)
        base64.b64encode(DATA_HUGE)
        base64.b64decode(B64_HUGE)
    return pyperf.perf_counter() - t0


# --- URL-safe Base64 (small only) ---

def bench_urlsafe_b64_small(loops):
    range_it = range(loops)
    t0 = pyperf.perf_counter()
    for _ in range_it:
        for _ in range(450):
            base64.urlsafe_b64encode(DATA_TINY)
            base64.urlsafe_b64decode(B64_URLSAFE_TINY)
        for _ in range(71):
            base64.urlsafe_b64encode(DATA_SMALL)
            base64.urlsafe_b64decode(B64_URLSAFE_SMALL)
        for _ in range(3):
            base64.urlsafe_b64encode(DATA_MEDIUM)
            base64.urlsafe_b64decode(B64_URLSAFE_MEDIUM)
        base64.urlsafe_b64encode(DATA_9K)
        base64.urlsafe_b64decode(B64_URLSAFE_9K_STR)
    return pyperf.perf_counter() - t0


# --- Base32 ---

def bench_b32_small(loops):
    range_it = range(loops)
    t0 = pyperf.perf_counter()
    for _ in range_it:
        for _ in range(450):
            base64.b32encode(DATA_TINY)
            base64.b32decode(B32_TINY)
        for _ in range(71):
            base64.b32encode(DATA_SMALL)
            base64.b32decode(B32_SMALL)
        for _ in range(3):
            base64.b32encode(DATA_MEDIUM)
            base64.b32decode(B32_MEDIUM)
        base64.b32encode(DATA_9K)
        base64.b32decode(B32_9K_STR)
    return pyperf.perf_counter() - t0


def bench_b32_large(loops):
    range_it = range(loops)
    t0 = pyperf.perf_counter()
    for _ in range_it:
        for _ in range(10):
            base64.b32encode(DATA_LARGE)
            base64.b32decode(B32_LARGE)
        base64.b32encode(DATA_HUGE)
        base64.b32decode(B32_HUGE)
    return pyperf.perf_counter() - t0


# --- Base16 ---

def bench_b16_small(loops):
    range_it = range(loops)
    t0 = pyperf.perf_counter()
    for _ in range_it:
        for _ in range(450):
            base64.b16encode(DATA_TINY)
            base64.b16decode(B16_TINY)
        for _ in range(71):
            base64.b16encode(DATA_SMALL)
            base64.b16decode(B16_SMALL)
        for _ in range(3):
            base64.b16encode(DATA_MEDIUM)
            base64.b16decode(B16_MEDIUM)
        base64.b16encode(DATA_9K)
        base64.b16decode(B16_9K_STR)
    return pyperf.perf_counter() - t0


def bench_b16_large(loops):
    range_it = range(loops)
    t0 = pyperf.perf_counter()
    for _ in range_it:
        for _ in range(10):
            base64.b16encode(DATA_LARGE)
            base64.b16decode(B16_LARGE)
        base64.b16encode(DATA_HUGE)
        base64.b16decode(B16_HUGE)
    return pyperf.perf_counter() - t0


# --- Ascii85 (includes wrapcol=76) ---

def bench_a85_small(loops):
    range_it = range(loops)
    t0 = pyperf.perf_counter()
    for _ in range_it:
        for _ in range(450):
            base64.a85encode(DATA_TINY)
            base64.a85encode(DATA_TINY, wrapcol=76)
            base64.a85decode(A85_TINY)
            base64.a85decode(A85_TINY)  # balance enc+dec weight
        for _ in range(71):
            base64.a85encode(DATA_SMALL)
            base64.a85encode(DATA_SMALL, wrapcol=76)
            base64.a85decode(A85_SMALL)
            base64.a85decode(A85_SMALL)  # balance enc+dec weight
        for _ in range(3):
            base64.a85encode(DATA_MEDIUM)
            base64.a85encode(DATA_MEDIUM, wrapcol=76)
            base64.a85decode(A85_MEDIUM)
            base64.a85decode(A85_MEDIUM)  # balance enc+dec weight
        base64.a85encode(DATA_9K)
        base64.a85encode(DATA_9K, wrapcol=76)
        base64.a85decode(A85_9K_STR)
        base64.a85decode(A85_9K_STR)  # balance enc+dec weight
    return pyperf.perf_counter() - t0


def bench_a85_large(loops):
    range_it = range(loops)
    t0 = pyperf.perf_counter()
    for _ in range_it:
        for _ in range(10):
            base64.a85encode(DATA_LARGE)
            base64.a85encode(DATA_LARGE, wrapcol=76)
            base64.a85decode(A85_LARGE)
            base64.a85decode(A85_LARGE)  # balance enc+dec weight
        base64.a85encode(DATA_HUGE)
        base64.a85encode(DATA_HUGE, wrapcol=76)
        base64.a85decode(A85_HUGE)
        base64.a85decode(A85_HUGE)  # balance enc+dec weight
    return pyperf.perf_counter() - t0


# --- Base85 ---

def bench_b85_small(loops):
    range_it = range(loops)
    t0 = pyperf.perf_counter()
    for _ in range_it:
        for _ in range(450):
            base64.b85encode(DATA_TINY)
            base64.b85decode(B85_TINY)
        for _ in range(71):
            base64.b85encode(DATA_SMALL)
            base64.b85decode(B85_SMALL)
        for _ in range(3):
            base64.b85encode(DATA_MEDIUM)
            base64.b85decode(B85_MEDIUM)
        base64.b85encode(DATA_9K)
        base64.b85decode(B85_9K_STR)
    return pyperf.perf_counter() - t0


def bench_b85_large(loops):
    range_it = range(loops)
    t0 = pyperf.perf_counter()
    for _ in range_it:
        for _ in range(10):
            base64.b85encode(DATA_LARGE)
            base64.b85decode(B85_LARGE)
        base64.b85encode(DATA_HUGE)
        base64.b85decode(B85_HUGE)
    return pyperf.perf_counter() - t0


if __name__ == "__main__":
    runner = pyperf.Runner()
    runner.metadata['description'] = "Benchmark base64 module encoding/decoding"

    runner.bench_time_func('base64_small', bench_b64_small)
    runner.bench_time_func('base64_large', bench_b64_large)

    runner.bench_time_func('urlsafe_base64_small', bench_urlsafe_b64_small)

    runner.bench_time_func('base32_small', bench_b32_small)
    runner.bench_time_func('base32_large', bench_b32_large)

    runner.bench_time_func('base16_small', bench_b16_small)
    runner.bench_time_func('base16_large', bench_b16_large)

    runner.bench_time_func('ascii85_small', bench_a85_small)
    runner.bench_time_func('ascii85_large', bench_a85_large)

    runner.bench_time_func('base85_small', bench_b85_small)
    runner.bench_time_func('base85_large', bench_b85_large)



"""
Benchmark recursive async generators implemented in python
by traversing a binary tree.

Author: Kumar Aditya
"""

from __future__ import annotations

from collections.abc import AsyncIterator

import pyperf


class Tree:
    def __init__(self, left: Tree | None, value: int, right: Tree | None) -> None:
        self.left = left
        self.value = value
        self.right = right

    async def __aiter__(self) -> AsyncIterator[int]:
        if self.left:
            async for i in self.left:
                yield i
        yield self.value
        if self.right:
            async for i in self.right:
                yield i


def tree(input: range) -> Tree | None:
    n = len(input)
    if n == 0:
        return None
    i = n // 2
    return Tree(tree(input[:i]), input[i], tree(input[i + 1:]))

async def bench_async_generators() -> None:
    async for _ in tree(range(100000)):
        pass

if __name__ == "__main__":
    runner = pyperf.Runner()
    runner.metadata['description'] = "Benchmark async generators"
    runner.bench_async_func('async_generators', bench_async_generators)



import glob
import os.path
import sys
import subprocess

import pyperf


if __name__ == "__main__":
    runner = pyperf.Runner()

    runner.metadata['description'] = "Performance of the Python 2to3 program"
    args = runner.parse_args()

    datadir = os.path.join(os.path.dirname(__file__), 'data', '2to3')
    pyfiles = glob.glob(os.path.join(datadir, '*.py.txt'))
    command = [sys.executable, "-m", "lib2to3", "-f", "all"] + pyfiles

    try:
        import lib2to3
    except ModuleNotFoundError:
        vendor = os.path.join(os.path.dirname(__file__), 'vendor')
        subprocess.run([sys.executable, "-m", "pip", "install", vendor], check=True)

    runner.bench_command('2to3', command)



# A utility library for getting information about a Python executable.
#
# This may be used as a script.

import importlib.util
import json
import os
import os.path
import sys
import sysconfig

INFO = {
    # sys
    "executable (sys)": "sys.executable",
    "executable (sys;realpath)": "executable_realpath",
    "prefix (sys)": "sys.prefix",
    "exec_prefix (sys)": "sys.exec_prefix",
    "stdlib_dir (sys)": "sys._stdlib_dir",
    "base_executable (sys)": "sys._base_executable",
    "base_prefix (sys)": "sys.base_prefix",
    "base_exec_prefix (sys)": "sys.base_exec_prefix",
    "version_str (sys)": "sys.version",
    "version_info (sys)": "sys.version_info",
    "hexversion (sys)": "sys.hexversion",
    "api_version (sys)": "sys.api_version",
    "implementation_name (sys)": "sys.implementation.name",
    "implementation_version (sys)": "sys.implementation.version",
    "platform (sys)": "sys.platform",
    # sysconfig
    "stdlib_dir (sysconfig)": "sysconfig.paths.stdlib",
    "is_dev (sysconfig)": "sysconfig.is_python_build",
    # other
    "base_executable": "base_executable",
    "stdlib_dir": "stdlib_dir",
    "pyc_magic_number": "pyc_magic_number",
    "is_venv": "is_venv",
}


def get_info(python=sys.executable):
    """Return an object with details about the given Python executable.

    Most of the details are grouped by their source.

    By default the current Python is used.
    """
    if python and python != sys.executable:
        # Run _pythoninfo.py to get the raw info.
        import subprocess

        argv = [python, __file__]
        try:
            text = subprocess.check_output(argv, encoding="utf-8")
        except subprocess.CalledProcessError:
            raise Exception(f"could not get info for {python or sys.executable}")
        data = _unjsonify_info(text)
    else:
        data = _get_current_info()
    return _build_info(data)


def _build_info(data):
    # Map the data into a new types.SimpleNamespace object.
    info = type(sys.implementation)()
    for key, value in data.items():
        try:
            field = INFO[key]
        except KeyError:
            raise NotImplementedError(repr(key))
        parent = info
        while "." in field:
            pname, _, field = field.partition(".")
            try:
                parent = getattr(parent, pname)
            except AttributeError:
                setattr(parent, pname, type(sys.implementation)())
                parent = getattr(parent, pname)
        setattr(parent, field, value)
    return info


def _get_current_info():
    is_venv = sys.prefix != sys.base_prefix
    base_executable = getattr(sys, "_base_executable", None)
    if is_venv:
        # XXX There is probably a bug related to venv, since
        # sys._base_executable should be different.
        if base_executable == sys.executable:
            # Indicate that we don't know.
            base_executable = None
    elif not base_executable:
        base_executable = sys.executable
    info = {
        # locations
        "executable (sys)": sys.executable,
        "executable (sys;realpath)": os.path.realpath(sys.executable),
        "prefix (sys)": sys.prefix,
        "exec_prefix (sys)": sys.exec_prefix,
        "stdlib_dir": os.path.dirname(os.__file__),
        "stdlib_dir (sys)": getattr(sys, "_stdlib_dir", None),
        "stdlib_dir (sysconfig)": (
            sysconfig.get_path("stdlib")
            if "stdlib" in sysconfig.get_path_names()
            else None
        ),
        # base locations
        "base_executable": base_executable,
        "base_executable (sys)": getattr(sys, "_base_executable", None),
        "base_prefix (sys)": sys.base_prefix,
        "base_exec_prefix (sys)": sys.base_exec_prefix,
        # version
        "version_str (sys)": sys.version,
        "version_info (sys)": sys.version_info,
        "hexversion (sys)": sys.hexversion,
        "api_version (sys)": sys.api_version,
        # implementation
        "implementation_name (sys)": sys.implementation.name,
        "implementation_version (sys)": sys.implementation.version,
        # build
        "is_dev (sysconfig)": sysconfig.is_python_build(),
        # host
        "platform (sys)": sys.platform,
        # virtual envs
        "is_venv": is_venv,
        # import system
        # importlib.util.MAGIC_NUMBER has been around since 3.5.
        "pyc_magic_number": importlib.util.MAGIC_NUMBER,
    }
    return info


def _jsonify_info(info):
    data = dict(info)
    if isinstance(data["pyc_magic_number"], bytes):
        data["pyc_magic_number"] = data["pyc_magic_number"].hex()
    return data


def _unjsonify_info(data):
    if isinstance(data, str):
        data = json.loads(data)
    info = dict(data)
    for key in ("version_info (sys)", "implementation_version (sys)"):
        if isinstance(info[key], list):
            # We would use type(sys.version_info) if it allowed it.
            info[key] = tuple(info[key])
    for key in ("pyc_magic_number",):
        if isinstance(info[key], str):
            info[key] = bytes.fromhex(data[key])
    return info


#######################################
# use as a script

if __name__ == "__main__":
    info = _get_current_info()
    data = _jsonify_info(info)
    json.dump(data, sys.stdout, indent=4)
    print()



"""
Render a template using Genshi module.
"""

import pyperf

from genshi.template import MarkupTemplate, NewTextTemplate


BIGTABLE_XML = """\
<table xmlns:py="http://genshi.edgewall.org/">
<tr py:for="row in table">
<td py:for="c in row.values()" py:content="c"/>
</tr>
</table>
"""

BIGTABLE_TEXT = """\
<table>
{% for row in table %}<tr>
{% for c in row.values() %}<td>$c</td>{% end %}
</tr>{% end %}
</table>
"""


def bench_genshi(loops, tmpl_cls, tmpl_str):
    tmpl = tmpl_cls(tmpl_str)
    table = [dict(a=1, b=2, c=3, d=4, e=5, f=6, g=7, h=8, i=9, j=10)
             for _ in range(1000)]
    range_it = range(loops)
    t0 = pyperf.perf_counter()

    for _ in range_it:
        stream = tmpl.generate(table=table)
        stream.render()

    return pyperf.perf_counter() - t0


def add_cmdline_args(cmd, args):
    if args.benchmark:
        cmd.append(args.benchmark)


BENCHMARKS = {
    'xml': (MarkupTemplate, BIGTABLE_XML),
    'text': (NewTextTemplate, BIGTABLE_TEXT),
}


if __name__ == "__main__":
    runner = pyperf.Runner(add_cmdline_args=add_cmdline_args)
    runner.metadata['description'] = "Render a template using Genshi module"
    runner.argparser.add_argument("benchmark", nargs='?',
                                  choices=sorted(BENCHMARKS))

    args = runner.parse_args()
    if args.benchmark:
        benchmarks = (args.benchmark,)
    else:
        benchmarks = sorted(BENCHMARKS)

    for bench in benchmarks:
        name = 'genshi_%s' % bench
        tmpl_cls, tmpl_str = BENCHMARKS[bench]
        runner.bench_time_func(name, bench_genshi, tmpl_cls, tmpl_str)



"""
Test the performance of isinstance() checks against runtime-checkable protocols.

For programmes that make extensive use of this feature,
these calls can easily become a bottleneck.
See https://github.com/python/cpython/issues/74690

The following situations all exercise different code paths
in typing._ProtocolMeta.__instancecheck__,
so each is tested in this benchmark:

  (1) Comparing an instance of a class that directly inherits
      from a protocol to that protocol.
  (2) Comparing an instance of a class that fulfils the interface
      of a protocol using instance attributes.
  (3) Comparing an instance of a class that fulfils the interface
      of a protocol using class attributes.
  (4) Comparing an instance of a class that fulfils the interface
      of a protocol using properties.

Protocols with callable and non-callable members also
exercise different code paths in _ProtocolMeta.__instancecheck__,
so are also tested separately.
"""

import time
from typing import Protocol, runtime_checkable

import pyperf


##################################################
# Protocols to call isinstance() against
##################################################


@runtime_checkable
class HasX(Protocol):
    """A runtime-checkable protocol with a single non-callable member"""
    x: int

@runtime_checkable
class HasManyAttributes(Protocol):
    """A runtime-checkable protocol with many non-callable members"""
    a: int
    b: int
    c: int
    d: int
    e: int

@runtime_checkable
class SupportsInt(Protocol):
    """A runtime-checkable protocol with a single callable member"""
    def __int__(self) -> int: ...

@runtime_checkable
class SupportsManyMethods(Protocol):
    """A runtime-checkable protocol with many callable members"""
    def one(self) -> int: ...
    def two(self) -> str: ...
    def three(self) -> bytes: ...
    def four(self) -> memoryview: ...
    def five(self) -> bytearray: ...

@runtime_checkable
class SupportsIntAndX(Protocol):
    """A runtime-checkable protocol with a mix of callable and non-callable members"""
    x: int
    def __int__(self) -> int: ...


##################################################
# Classes for comparing against the protocols
##################################################


class Empty:
    """Empty class with no attributes"""

class PropertyX:
    """Class with a property x"""
    @property
    def x(self) -> int: return 42

class HasIntMethod:
    """Class with an __int__ method"""
    def __int__(self): return 42

class HasManyMethods:
    """Class with many methods"""
    def one(self): return 42
    def two(self): return "42"
    def three(self): return b"42"
    def four(self): return memoryview(b"42")
    def five(self): return bytearray(b"42")

class PropertyXWithInt:
    """Class with a property x and an __int__ method"""
    @property
    def x(self) -> int: return 42
    def __int__(self): return 42

class ClassVarX:
    """Class with a ClassVar x"""
    x = 42

class ClassVarXWithInt:
    """Class with a ClassVar x and an __int__ method"""
    x = 42
    def __int__(self): return 42

class InstanceVarX:
    """Class with an instance var x"""
    def __init__(self):
        self.x = 42

class ManyInstanceVars:
    """Class with many instance vars"""
    def __init__(self):
        for attr in 'abcde':
            setattr(self, attr, 42)

class InstanceVarXWithInt:
    """Class with an instance var x and an __int__ method"""
    def __init__(self):
        self.x = 42
    def __int__(self):
        return 42

class NominalX(HasX):
    """Class that explicitly subclasses HasX"""
    def __init__(self):
        self.x = 42

class NominalSupportsInt(SupportsInt):
    """Class that explicitly subclasses SupportsInt"""
    def __int__(self):
        return 42

class NominalXWithInt(SupportsIntAndX):
    """Class that explicitly subclasses NominalXWithInt"""
    def __init__(self):
        self.x = 42


##################################################
# Time to test the performance of isinstance()!
##################################################


def bench_protocols(loops: int) -> float:
    protocols = [
        HasX, HasManyAttributes, SupportsInt, SupportsManyMethods, SupportsIntAndX
    ]
    instances = [
        cls()
        for cls in (
            Empty, PropertyX, HasIntMethod, HasManyMethods, PropertyXWithInt,
            ClassVarX, ClassVarXWithInt, InstanceVarX, ManyInstanceVars,
            InstanceVarXWithInt, NominalX, NominalSupportsInt, NominalXWithInt
        )
    ]

    t0 = time.perf_counter()

    for _ in range(loops):
        for protocol in protocols:
            for instance in instances:
                isinstance(instance, protocol)

    return time.perf_counter() - t0


if __name__ == "__main__":
    runner = pyperf.Runner()
    runner.metadata["description"] = (
        "Test the performance of isinstance() checks "
        "against runtime-checkable protocols"
    )
    runner.bench_time_func("typing_runtime_protocols", bench_protocols)



"""
Artificial, floating point-heavy benchmark originally used by Factor.
"""
import pyperf

from math import sin, cos, sqrt


POINTS = 100000


class Point(object):
    __slots__ = ('x', 'y', 'z')

    def __init__(self, i):
        self.x = x = sin(i)
        self.y = cos(i) * 3
        self.z = (x * x) / 2

    def __repr__(self):
        return "<Point: x=%s, y=%s, z=%s>" % (self.x, self.y, self.z)

    def normalize(self):
        x = self.x
        y = self.y
        z = self.z
        norm = sqrt(x * x + y * y + z * z)
        self.x /= norm
        self.y /= norm
        self.z /= norm

    def maximize(self, other):
        self.x = self.x if self.x > other.x else other.x
        self.y = self.y if self.y > other.y else other.y
        self.z = self.z if self.z > other.z else other.z
        return self


def maximize(points):
    next = points[0]
    for p in points[1:]:
        next = next.maximize(p)
    return next


def benchmark(n):
    points = [None] * n
    for i in range(n):
        points[i] = Point(i)
    for p in points:
        p.normalize()
    return maximize(points)


if __name__ == "__main__":
    runner = pyperf.Runner()
    runner.metadata['description'] = "Float benchmark"

    points = POINTS
    runner.bench_func('float', benchmark, points)



"""
Go board game
"""
import math
import random

import pyperf


SIZE = 9
GAMES = 200
KOMI = 7.5
EMPTY, WHITE, BLACK = 0, 1, 2
SHOW = {EMPTY: '.', WHITE: 'o', BLACK: 'x'}
PASS = -1
MAXMOVES = SIZE * SIZE * 3
TIMESTAMP = 0
MOVES = 0


def to_pos(x, y):
    return y * SIZE + x


def to_xy(pos):
    y, x = divmod(pos, SIZE)
    return x, y


class Square:

    def __init__(self, board, pos):
        self.board = board
        self.pos = pos
        self.timestamp = TIMESTAMP
        self.removestamp = TIMESTAMP
        self.zobrist_strings = [random.randrange(9223372036854775807)
                                for i in range(3)]

    def set_neighbours(self):
        x, y = self.pos % SIZE, self.pos // SIZE
        self.neighbours = []
        for dx, dy in [(-1, 0), (1, 0), (0, -1), (0, 1)]:
            newx, newy = x + dx, y + dy
            if 0 <= newx < SIZE and 0 <= newy < SIZE:
                self.neighbours.append(self.board.squares[to_pos(newx, newy)])

    def move(self, color):
        global TIMESTAMP, MOVES
        TIMESTAMP += 1
        MOVES += 1
        self.board.zobrist.update(self, color)
        self.color = color
        self.reference = self
        self.ledges = 0
        self.used = True
        for neighbour in self.neighbours:
            neighcolor = neighbour.color
            if neighcolor == EMPTY:
                self.ledges += 1
            else:
                neighbour_ref = neighbour.find(update=True)
                if neighcolor == color:
                    if neighbour_ref.reference.pos != self.pos:
                        self.ledges += neighbour_ref.ledges
                        neighbour_ref.reference = self
                    self.ledges -= 1
                else:
                    neighbour_ref.ledges -= 1
                    if neighbour_ref.ledges == 0:
                        neighbour.remove(neighbour_ref)
        self.board.zobrist.add()

    def remove(self, reference, update=True):
        self.board.zobrist.update(self, EMPTY)
        self.removestamp = TIMESTAMP
        if update:
            self.color = EMPTY
            self.board.emptyset.add(self.pos)
#            if color == BLACK:
#                self.board.black_dead += 1
#            else:
#                self.board.white_dead += 1
        for neighbour in self.neighbours:
            if neighbour.color != EMPTY and neighbour.removestamp != TIMESTAMP:
                neighbour_ref = neighbour.find(update)
                if neighbour_ref.pos == reference.pos:
                    neighbour.remove(reference, update)
                else:
                    if update:
                        neighbour_ref.ledges += 1

    def find(self, update=False):
        reference = self.reference
        if reference.pos != self.pos:
            reference = reference.find(update)
            if update:
                self.reference = reference
        return reference

    def __repr__(self):
        return repr(to_xy(self.pos))


class EmptySet:

    def __init__(self, board):
        self.board = board
        self.empties = list(range(SIZE * SIZE))
        self.empty_pos = list(range(SIZE * SIZE))

    def random_choice(self):
        choices = len(self.empties)
        while choices:
            i = int(random.random() * choices)
            pos = self.empties[i]
            if self.board.useful(pos):
                return pos
            choices -= 1
            self.set(i, self.empties[choices])
            self.set(choices, pos)
        return PASS

    def add(self, pos):
        self.empty_pos[pos] = len(self.empties)
        self.empties.append(pos)

    def remove(self, pos):
        self.set(self.empty_pos[pos], self.empties[len(self.empties) - 1])
        self.empties.pop()

    def set(self, i, pos):
        self.empties[i] = pos
        self.empty_pos[pos] = i


class ZobristHash:

    def __init__(self, board):
        self.board = board
        self.hash_set = set()
        self.hash = 0
        for square in self.board.squares:
            self.hash ^= square.zobrist_strings[EMPTY]
        self.hash_set.clear()
        self.hash_set.add(self.hash)

    def update(self, square, color):
        self.hash ^= square.zobrist_strings[square.color]
        self.hash ^= square.zobrist_strings[color]

    def add(self):
        self.hash_set.add(self.hash)

    def dupe(self):
        return self.hash in self.hash_set


class Board:

    def __init__(self):
        self.squares = [Square(self, pos) for pos in range(SIZE * SIZE)]
        for square in self.squares:
            square.set_neighbours()
        self.reset()

    def reset(self):
        for square in self.squares:
            square.color = EMPTY
            square.used = False
        self.emptyset = EmptySet(self)
        self.zobrist = ZobristHash(self)
        self.color = BLACK
        self.finished = False
        self.lastmove = -2
        self.history = []
        self.white_dead = 0
        self.black_dead = 0

    def move(self, pos):
        square = self.squares[pos]
        if pos != PASS:
            square.move(self.color)
            self.emptyset.remove(square.pos)
        elif self.lastmove == PASS:
            self.finished = True
        if self.color == BLACK:
            self.color = WHITE
        else:
            self.color = BLACK
        self.lastmove = pos
        self.history.append(pos)

    def random_move(self):
        return self.emptyset.random_choice()

    def useful_fast(self, square):
        if not square.used:
            for neighbour in square.neighbours:
                if neighbour.color == EMPTY:
                    return True
        return False

    def useful(self, pos):
        global TIMESTAMP
        TIMESTAMP += 1
        square = self.squares[pos]
        if self.useful_fast(square):
            return True
        old_hash = self.zobrist.hash
        self.zobrist.update(square, self.color)
        empties = opps = weak_opps = neighs = weak_neighs = 0
        for neighbour in square.neighbours:
            neighcolor = neighbour.color
            if neighcolor == EMPTY:
                empties += 1
                continue
            neighbour_ref = neighbour.find()
            if neighbour_ref.timestamp != TIMESTAMP:
                if neighcolor == self.color:
                    neighs += 1
                else:
                    opps += 1
                neighbour_ref.timestamp = TIMESTAMP
                neighbour_ref.temp_ledges = neighbour_ref.ledges
            neighbour_ref.temp_ledges -= 1
            if neighbour_ref.temp_ledges == 0:
                if neighcolor == self.color:
                    weak_neighs += 1
                else:
                    weak_opps += 1
                    neighbour_ref.remove(neighbour_ref, update=False)
        dupe = self.zobrist.dupe()
        self.zobrist.hash = old_hash
        strong_neighs = neighs - weak_neighs
        strong_opps = opps - weak_opps
        return not dupe and \
            (empties or weak_opps or (strong_neighs and (strong_opps or weak_neighs)))

    def useful_moves(self):
        return [pos for pos in self.emptyset.empties if self.useful(pos)]

    def replay(self, history):
        for pos in history:
            self.move(pos)

    def score(self, color):
        if color == WHITE:
            count = KOMI + self.black_dead
        else:
            count = self.white_dead
        for square in self.squares:
            squarecolor = square.color
            if squarecolor == color:
                count += 1
            elif squarecolor == EMPTY:
                surround = 0
                for neighbour in square.neighbours:
                    if neighbour.color == color:
                        surround += 1
                if surround == len(square.neighbours):
                    count += 1
        return count

    def check(self):
        for square in self.squares:
            if square.color == EMPTY:
                continue

            members1 = set([square])
            changed = True
            while changed:
                changed = False
                for member in members1.copy():
                    for neighbour in member.neighbours:
                        if neighbour.color == square.color and neighbour not in members1:
                            changed = True
                            members1.add(neighbour)
            ledges1 = 0
            for member in members1:
                for neighbour in member.neighbours:
                    if neighbour.color == EMPTY:
                        ledges1 += 1

            root = square.find()

            # print 'members1', square, root, members1
            # print 'ledges1', square, ledges1

            members2 = set()
            for square2 in self.squares:
                if square2.color != EMPTY and square2.find() == root:
                    members2.add(square2)

            ledges2 = root.ledges
            # print 'members2', square, root, members1
            # print 'ledges2', square, ledges2

            assert members1 == members2
            assert ledges1 == ledges2, ('ledges differ at %r: %d %d' % (
                square, ledges1, ledges2))

            set(self.emptyset.empties)

            empties2 = set()
            for square in self.squares:
                if square.color == EMPTY:
                    empties2.add(square.pos)

    def __repr__(self):
        result = []
        for y in range(SIZE):
            start = to_pos(0, y)
            result.append(''.join(
                [SHOW[square.color] + ' ' for square in self.squares[start:start + SIZE]]))
        return '\n'.join(result)


class UCTNode:

    def __init__(self):
        self.bestchild = None
        self.pos = -1
        self.wins = 0
        self.losses = 0
        self.pos_child = [None for x in range(SIZE * SIZE)]
        self.parent = None

    def play(self, board):
        """ uct tree search """
        color = board.color
        node = self
        path = [node]
        while True:
            pos = node.select(board)
            if pos == PASS:
                break
            board.move(pos)
            child = node.pos_child[pos]
            if not child:
                child = node.pos_child[pos] = UCTNode()
                child.unexplored = board.useful_moves()
                child.pos = pos
                child.parent = node
                path.append(child)
                break
            path.append(child)
            node = child
        self.random_playout(board)
        self.update_path(board, color, path)

    def select(self, board):
        """ select move; unexplored children first, then according to uct value """
        if self.unexplored:
            i = random.randrange(len(self.unexplored))
            pos = self.unexplored[i]
            self.unexplored[i] = self.unexplored[len(self.unexplored) - 1]
            self.unexplored.pop()
            return pos
        elif self.bestchild:
            return self.bestchild.pos
        else:
            return PASS

    def random_playout(self, board):
        """ random play until both players pass """
        for x in range(MAXMOVES):  # XXX while not self.finished?
            if board.finished:
                break
            board.move(board.random_move())

    def update_path(self, board, color, path):
        """ update win/loss count along path """
        wins = board.score(BLACK) >= board.score(WHITE)
        for node in path:
            if color == BLACK:
                color = WHITE
            else:
                color = BLACK
            if wins == (color == BLACK):
                node.wins += 1
            else:
                node.losses += 1
            if node.parent:
                node.parent.bestchild = node.parent.best_child()

    def score(self):
        winrate = self.wins / float(self.wins + self.losses)
        parentvisits = self.parent.wins + self.parent.losses
        if not parentvisits:
            return winrate
        nodevisits = self.wins + self.losses
        return winrate + math.sqrt((math.log(parentvisits)) / (5 * nodevisits))

    def best_child(self):
        maxscore = -1
        maxchild = None
        for child in self.pos_child:
            if child and child.score() > maxscore:
                maxchild = child
                maxscore = child.score()
        return maxchild

    def best_visited(self):
        maxvisits = -1
        maxchild = None
        for child in self.pos_child:
            #            if child:
            # print to_xy(child.pos), child.wins, child.losses, child.score()
            if child and (child.wins + child.losses) > maxvisits:
                maxvisits, maxchild = (child.wins + child.losses), child
        return maxchild


# def user_move(board):
#     while True:
#         text = input('?').strip()
#         if text == 'p':
#             return PASS
#         if text == 'q':
#             raise EOFError
#         try:
#             x, y = [int(i) for i in text.split()]
#         except ValueError:
#             continue
#         if not (0 <= x < SIZE and 0 <= y < SIZE):
#             continue
#         pos = to_pos(x, y)
#         if board.useful(pos):
#             return pos


def computer_move(board):
    pos = board.random_move()
    if pos == PASS:
        return PASS
    tree = UCTNode()
    tree.unexplored = board.useful_moves()
    nboard = Board()
    for game in range(GAMES):
        node = tree
        nboard.reset()
        nboard.replay(board.history)
        node.play(nboard)
    return tree.best_visited().pos


def versus_cpu():
    random.seed(1)
    board = Board()
    return computer_move(board)


if __name__ == "__main__":
    runner = pyperf.Runner()
    runner.metadata['description'] = "Test the performance of the Go benchmark"
    runner.bench_func('go', versus_cpu)



import pyperf
import gc

N_LEVELS = 1000


def create_recursive_containers(n_levels):

    current_list = []
    for n in range(n_levels):
        new_list = [None] * n
        for index in range(n):
            new_list[index] = current_list
        current_list = new_list

    return current_list


def benchamark_collection(loops, n_levels):
    total_time = 0
    all_cycles = create_recursive_containers(n_levels)
    for _ in range(loops):
        gc.collect()
        # Main loop to measure
        t0 = pyperf.perf_counter()
        collected = gc.collect()
        total_time += pyperf.perf_counter() - t0

        assert collected is None or collected == 0

    return total_time


if __name__ == "__main__":
    runner = pyperf.Runner()
    runner.metadata["description"] = "GC traversal benchmark"
    runner.bench_time_func("gc_traversal", benchamark_collection, N_LEVELS)



"""
Benchmark the Dask scheduler running a large number of simple jobs.

Author: Matt Rocklin, Michael Droettboom
"""

from dask.distributed import Client, Worker, Scheduler, wait
from dask import distributed

import pyperf


def inc(x):
    return x + 1


async def benchmark():
    async with Scheduler() as scheduler:
        async with Worker(scheduler.address):
            async with Client(scheduler.address, asynchronous=True) as client:

                futures = client.map(inc, range(100))
                for _ in range(10):
                    futures = client.map(inc, futures)

                await wait(futures)


if __name__ == "__main__":
    runner = pyperf.Runner()
    runner.metadata['description'] = "Benchmark dask"
    runner.bench_async_func('dask', benchmark)



import os
import os.path
import sys

import pyperformance

from . import _pip, _utils, _venv

REQUIREMENTS_FILE = os.path.join(
    os.path.dirname(__file__), "requirements", "requirements.txt"
)
PYPERF_OPTIONAL = ["psutil"]


class Requirements(object):
    @classmethod
    def from_file(cls, filename):
        self = cls()
        self._add_from_file(filename)
        return self

    @classmethod
    def from_benchmarks(cls, benchmarks):
        self = cls()
        for bench in benchmarks or ():
            filename = bench.requirements_lockfile
            self._add_from_file(filename)
        return self

    def __init__(self):
        # if pip or setuptools is updated:
        # .github/workflows/main.yml should be updated as well

        # requirements
        self.specs = []

    def __len__(self):
        return len(self.specs)

    def __iter__(self):
        for spec in self.specs:
            yield spec

    def _add_from_file(self, filename):
        if not os.path.exists(filename):
            return
        for line in _utils.iter_clean_lines(filename):
            fullpath = os.path.join(os.path.dirname(filename), line.strip())
            if os.path.isfile(fullpath):
                self._add(fullpath)
            else:
                self._add(line)

    def _add(self, line):
        self.specs.append(line)

    def get(self, name):
        for req in self.specs:
            if _pip.get_pkg_name(req) == name:
                return req
        return None


# This is used by the hg_startup benchmark.
def get_venv_program(program):
    bin_path = os.path.dirname(sys.executable)
    bin_path = os.path.realpath(bin_path)

    if not os.path.isabs(bin_path):
        print("ERROR: Python executable path is not absolute: %s" % sys.executable)
        sys.exit(1)

    if not os.path.exists(os.path.join(bin_path, "activate")):
        print(
            "ERROR: Unable to get the virtual environment of "
            "the Python executable %s" % sys.executable
        )
        sys.exit(1)

    if os.name == "nt":
        path = os.path.join(bin_path, program)
    else:
        path = os.path.join(bin_path, program)

    if not os.path.exists(path):
        print(
            "ERROR: Unable to get the program %r "
            "from the virtual environment %r" % (program, bin_path)
        )
        sys.exit(1)

    return path


NECESSARY_ENV_VARS = {
    "nt": [
        "ALLUSERSPROFILE",
        "APPDATA",
        "COMPUTERNAME",
        "ComSpec",
        "CommonProgramFiles",
        "CommonProgramFiles(x86)",
        "CommonProgramW6432",
        "HOMEDRIVE",
        "HOMEPATH",
        "LOCALAPPDATA",
        "NUMBER_OF_PROCESSORS",
        "OS",
        "PATHEXT",
        "PROCESSOR_ARCHITECTURE",
        "PROCESSOR_IDENTIFIER",
        "PROCESSOR_LEVEL",
        "PROCESSOR_REVISION",
        "Path",
        "ProgramData",
        "ProgramFiles",
        "ProgramFiles(x86)",
        "ProgramW6432",
        "SystemDrive",
        "SystemRoot",
        "TEMP",
        "TMP",
        "USERDNSDOMAIN",
        "USERDOMAIN",
        "USERDOMAIN_ROAMINGPROFILE",
        "USERNAME",
        "USERPROFILE",
        "windir",
    ],
}
NECESSARY_ENV_VARS_DEFAULT = [
    "HOME",
    "PATH",
]


def _get_envvars(inherit=None, osname=None):
    # Restrict the env we use.
    try:
        necessary = NECESSARY_ENV_VARS[osname or os.name]
    except KeyError:
        necessary = NECESSARY_ENV_VARS_DEFAULT
    copy_env = list(necessary)
    if inherit:
        copy_env.extend(inherit)

    env = {}
    for name in copy_env:
        if name in os.environ:
            env[name] = os.environ[name]
    return env


class VenvForBenchmarks(_venv.VirtualEnvironment):
    @classmethod
    def create(
        cls,
        root=None,
        python=None,
        *,
        inherit_environ=None,
        upgrade=False,
    ):
        env = _get_envvars(inherit_environ)
        self = super().create(root, python, env=env, withpip=False)
        self.inherit_environ = inherit_environ

        try:
            self.ensure_pip(upgrade=upgrade)
        except BaseException:
            _utils.safe_rmtree(self.root)
            raise

        # Display the pip version
        _pip.run_pip("--version", python=self.python, env=self._env)

        return self

    @classmethod
    def ensure(
        cls, root, python=None, *, inherit_environ=None, upgrade=False, **kwargs
    ):
        exists = _venv.venv_exists(root)
        if upgrade == "oncreate":
            upgrade = not exists
        elif upgrade == "onexists":
            upgrade = exists
        elif isinstance(upgrade, str):
            raise NotImplementedError(upgrade)

        if exists:
            self = super().ensure(root)
            self.inherit_environ = inherit_environ
            if upgrade:
                self.upgrade_pip()
            else:
                self.ensure_pip(upgrade=False)
            return self
        else:
            return cls.create(
                root, python, inherit_environ=inherit_environ, upgrade=upgrade, **kwargs
            )

    def __init__(self, root, *, base=None, inherit_environ=None):
        super().__init__(root, base=base)
        self.inherit_environ = inherit_environ or None

    @property
    def _env(self):
        # Restrict the env we use.
        return _get_envvars(self.inherit_environ)

    def install_pyperformance(self):
        print("installing pyperformance in the venv at %s" % self.root)
        # Install pyperformance inside the virtual environment.
        if pyperformance.is_dev():
            basereqs = Requirements.from_file(REQUIREMENTS_FILE)
            self.ensure_reqs(basereqs)
            if basereqs.get("pyperf"):
                self._install_pyperf_optional_dependencies()

            root_dir = os.path.dirname(pyperformance.PKG_ROOT)
            ec, _, _ = _pip.install_editable(
                root_dir,
                python=self.info,
                env=self._env,
            )
            if ec != 0:
                raise _venv.RequirementsInstallationFailedError(root_dir)
        else:
            version = pyperformance.__version__
            self.ensure_reqs([f"pyperformance=={version}"])
            self._install_pyperf_optional_dependencies()

    def _install_pyperf_optional_dependencies(self):
        for req in PYPERF_OPTIONAL:
            try:
                self.ensure_reqs([req])
            except _venv.RequirementsInstallationFailedError:
                print("WARNING: failed to install %s" % req)
                pass

    def ensure_reqs(self, requirements=None):
        # parse requirements
        bench = None
        if requirements is None:
            requirements = Requirements()
        elif hasattr(requirements, "requirements_lockfile"):
            bench = requirements
            requirements = Requirements.from_benchmarks([bench])

        # Every benchmark must depend on pyperf.
        if bench is not None and not requirements.get("pyperf"):
            basereqs = Requirements.from_file(REQUIREMENTS_FILE)
            pyperf_req = basereqs.get("pyperf")
            if not pyperf_req:
                raise NotImplementedError
            requirements.specs.append(pyperf_req)
            # XXX what about psutil?

        if not requirements:
            print("(nothing to install)")
        else:
            # install requirements
            super().ensure_reqs(
                *requirements,
                upgrade=False,
            )

            if bench is not None:
                self._install_pyperf_optional_dependencies()

        # Dump the package list and their versions: pip freeze
        _pip.run_pip("freeze", python=self.python, env=self._env)

        return requirements



__all__ = [
    "load_metadata",
]


import os.path

from . import _benchmark, _pyproject_toml, _utils

METADATA = "pyproject.toml"
DEPENDENCIES = "requirements.in"
REQUIREMENTS = "requirements.txt"
DATA = "data"
RUN = "run_benchmark.py"

PEP_621_FIELDS = {
    "name": None,
    "version": None,
    "requires-python": "python",
    "dependencies": None,
    #'optional-dependencies': '',
    #'urls': '',
}
TOOL_FIELDS = {
    #'inherits': None,
    "metafile": None,
    "name": None,
    "tags": None,
    "datadir": None,
    "runscript": None,
    "extra_opts": None,
}


# class BenchmarkMetadata:
#    spec
#    base
#    metafile
#    tags
#    python
#    dependencies  # (from requirements.in)
#    requirements  # (from lock file or requirements.txt)
#    datadir
#    runscript
#    extra_opts


def load_metadata(metafile, defaults=None):
    if isinstance(metafile, str):
        name, rootdir = _name_from_filename(metafile)
        data, filename = _pyproject_toml.load_pyproject_toml(
            metafile,
            name=name or None,
            requirefiles=False,
        )
    else:
        text = metafile.read()
        filename = metafile.name
        name, rootdir = _name_from_filename(filename)
        data = _pyproject_toml.parse_pyproject_toml(
            text,
            rootdir,
            name,
            requirefiles=False,
        )
    project = data.get("project")
    tool = data.get("tool", {}).get("pyperformance", {})

    defaults = _ensure_defaults(defaults, rootdir)
    base, basefile = _resolve_base(
        tool.get("inherits"),  # XXX Pop it?
        project,
        filename,
        defaults,
    )
    top = _resolve(project or {}, tool, filename)
    merged = _merge_metadata(top, base, defaults)

    if not merged.get("name"):
        raise ValueError("missing benchmark name")
    if not merged.get("version"):
        print("====================")
        from pprint import pprint

        print("top:")
        pprint(top)
        print("base:")
        pprint(base)
        print("defaults:")
        pprint(defaults)
        print("merged:")
        pprint(merged)
        print("====================")
        raise ValueError("missing benchmark version")

    metafile = merged.pop("metafile")
    merged["spec"] = _benchmark.BenchmarkSpec(
        merged.pop("name"),
        merged.pop("version"),
        # XXX Should we leave this (origin) blank?
        metafile,
    )
    if basefile:
        merged["base"] = basefile

    return merged, filename


#######################################
# internal implementation


def _name_from_filename(metafile):
    rootdir, basename = os.path.split(metafile)
    if basename == "pyproject.toml":
        dirname = os.path.dirname(rootdir)
        name = dirname[3:] if dirname.startswith("bm_") else None
    elif basename.startswith("bm_") and basename.endswith(".toml"):
        name = basename[3:-5]
    else:
        name = None
    return name, rootdir


def _ensure_defaults(defaults, rootdir):
    if not defaults:
        defaults = {}

    if not defaults.get("datadir"):
        datadir = os.path.join(rootdir, DATA)
        if os.path.isdir(datadir):
            defaults["datadir"] = datadir

    if not defaults.get("runscript"):
        runscript = os.path.join(rootdir, RUN)
        if os.path.isfile(runscript):
            defaults["runscript"] = runscript

    return defaults


def _resolve_base(metabase, project, filename, defaults, *, minimalwithbase=False):
    rootdir, basename = os.path.split(filename)

    if not metabase:
        if basename == "pyproject.toml":
            return None, None
        elif not (basename.startswith("bm_") and basename.endswith(".toml")):
            return None, None
        elif not os.path.basename(rootdir).startswith("bm_"):
            return None, None
        else:
            metabase = os.path.join(rootdir, "pyproject.toml")
            if not os.path.isfile(metabase):
                return None, None

    if project is not None and minimalwithbase:
        unexpected = set(project) - {"name", "dynamic", "dependencies"}
        if unexpected:
            raise ValueError(
                f'[project] should be minimal if "inherits" is provided, got extra {sorted(unexpected)}'
            )

    if metabase == "..":
        metabase = os.path.join(
            os.path.dirname(rootdir),
            "base.toml",
        )
    if metabase == filename:
        raise Exception("circular")

    if not os.path.isabs(metabase):
        metabase = os.path.join(rootdir, metabase)
        if metabase == filename:
            raise Exception("circular")

    defaults = dict(defaults, name="_base_")
    return load_metadata(metabase, defaults)


def _resolve(project, tool, filename):
    resolved = {
        "metafile": filename,
    }

    rootdir = os.path.dirname(filename)
    for field, target in TOOL_FIELDS.items():
        if target is None:
            target = field
        if not resolved.get(target):
            value = tool.get(field)
            if value is not None:
                resolved[target] = _resolve_value(field, value, rootdir)

    for field, target in PEP_621_FIELDS.items():
        if target is None:
            target = field
        if field == "url":
            raise NotImplementedError
        elif not resolved.get(target):
            value = project.get(field)
            if value is not None:
                resolved[target] = value

    return resolved


def _resolve_value(field, value, rootdir):
    if field == "name":
        _utils.check_name(value, allownumeric=True)
    elif field == "metafile":
        assert False, "unreachable"
    elif field == "tags":
        if isinstance(value, str):
            value = value.replace(",", " ").split()
        for tag in value:
            _utils.check_name(tag)
            if tag == "all":
                raise ValueError("Invalid tag 'all'")
            elif tag == "":
                raise ValueError("Invalid empty tag")
    elif field == "datadir":
        if not os.path.isabs(value):
            value = os.path.join(rootdir, value)
        _utils.check_dir(value)
    elif field == "runscript":
        if not os.path.isabs(value):
            value = os.path.join(rootdir, value)
        _utils.check_file(value)
    elif field == "extra_opts":
        if isinstance(value, str):
            raise TypeError(f"extra_opts should be a list of strings, got {value!r}")
        for opt in value:
            if not opt or not isinstance(opt, str):
                raise TypeError(
                    f"extra_opts should be a list of strings, got {value!r}"
                )
    else:
        raise NotImplementedError(field)
    return value


def _merge_metadata(*tiers):
    merged = {}
    for data in tiers:
        if not data:
            continue
        for field, value in data.items():
            if field == "spec":
                field = "version"
                value = value.version
            if merged.get(field):
                # XXX Merge containers?
                continue
            if value or isinstance(value, int):
                merged[field] = value
    return merged



"""
Solver of Hexiom board game.

Benchmark from Laurent Vaucher.

Source: https://github.com/slowfrog/hexiom : hexiom2.py, level36.txt

(Main function tweaked by Armin Rigo.)
"""

import io

import pyperf

# 2016-07-07: CPython 3.6 takes ~25 ms to solve the board level 25
DEFAULT_LEVEL = 25


##################################
class Dir(object):

    def __init__(self, x, y):
        self.x = x
        self.y = y


DIRS = [Dir(1, 0),
        Dir(-1, 0),
        Dir(0, 1),
        Dir(0, -1),
        Dir(1, 1),
        Dir(-1, -1)]

EMPTY = 7

##################################


class Done(object):
    MIN_CHOICE_STRATEGY = 0
    MAX_CHOICE_STRATEGY = 1
    HIGHEST_VALUE_STRATEGY = 2
    FIRST_STRATEGY = 3
    MAX_NEIGHBORS_STRATEGY = 4
    MIN_NEIGHBORS_STRATEGY = 5

    def __init__(self, count, empty=False):
        self.count = count
        self.cells = None if empty else [
            [0, 1, 2, 3, 4, 5, 6, EMPTY] for i in range(count)]

    def clone(self):
        ret = Done(self.count, True)
        ret.cells = [self.cells[i][:] for i in range(self.count)]
        return ret

    def __getitem__(self, i):
        return self.cells[i]

    def set_done(self, i, v):
        self.cells[i] = [v]

    def already_done(self, i):
        return len(self.cells[i]) == 1

    def remove(self, i, v):
        if v in self.cells[i]:
            self.cells[i].remove(v)
            return True
        else:
            return False

    def remove_all(self, v):
        for i in range(self.count):
            self.remove(i, v)

    def remove_unfixed(self, v):
        changed = False
        for i in range(self.count):
            if not self.already_done(i):
                if self.remove(i, v):
                    changed = True
        return changed

    def filter_tiles(self, tiles):
        for v in range(8):
            if tiles[v] == 0:
                self.remove_all(v)

    def next_cell_min_choice(self):
        minlen = 10
        mini = -1
        for i in range(self.count):
            if 1 < len(self.cells[i]) < minlen:
                minlen = len(self.cells[i])
                mini = i
        return mini

    def next_cell_max_choice(self):
        maxlen = 1
        maxi = -1
        for i in range(self.count):
            if maxlen < len(self.cells[i]):
                maxlen = len(self.cells[i])
                maxi = i
        return maxi

    def next_cell_highest_value(self):
        maxval = -1
        maxi = -1
        for i in range(self.count):
            if (not self.already_done(i)):
                maxvali = max(k for k in self.cells[i] if k != EMPTY)
                if maxval < maxvali:
                    maxval = maxvali
                    maxi = i
        return maxi

    def next_cell_first(self):
        for i in range(self.count):
            if (not self.already_done(i)):
                return i
        return -1

    def next_cell_max_neighbors(self, pos):
        maxn = -1
        maxi = -1
        for i in range(self.count):
            if not self.already_done(i):
                cells_around = pos.hex.get_by_id(i).links
                n = sum(1 if (self.already_done(nid) and (self[nid][0] != EMPTY)) else 0
                        for nid in cells_around)
                if n > maxn:
                    maxn = n
                    maxi = i
        return maxi

    def next_cell_min_neighbors(self, pos):
        minn = 7
        mini = -1
        for i in range(self.count):
            if not self.already_done(i):
                cells_around = pos.hex.get_by_id(i).links
                n = sum(1 if (self.already_done(nid) and (self[nid][0] != EMPTY)) else 0
                        for nid in cells_around)
                if n < minn:
                    minn = n
                    mini = i
        return mini

    def next_cell(self, pos, strategy=HIGHEST_VALUE_STRATEGY):
        if strategy == Done.HIGHEST_VALUE_STRATEGY:
            return self.next_cell_highest_value()
        elif strategy == Done.MIN_CHOICE_STRATEGY:
            return self.next_cell_min_choice()
        elif strategy == Done.MAX_CHOICE_STRATEGY:
            return self.next_cell_max_choice()
        elif strategy == Done.FIRST_STRATEGY:
            return self.next_cell_first()
        elif strategy == Done.MAX_NEIGHBORS_STRATEGY:
            return self.next_cell_max_neighbors(pos)
        elif strategy == Done.MIN_NEIGHBORS_STRATEGY:
            return self.next_cell_min_neighbors(pos)
        else:
            raise Exception("Wrong strategy: %d" % strategy)

##################################


class Node(object):

    def __init__(self, pos, id, links):
        self.pos = pos
        self.id = id
        self.links = links

##################################


class Hex(object):

    def __init__(self, size):
        self.size = size
        self.count = 3 * size * (size - 1) + 1
        self.nodes_by_id = self.count * [None]
        self.nodes_by_pos = {}
        id = 0
        for y in range(size):
            for x in range(size + y):
                pos = (x, y)
                node = Node(pos, id, [])
                self.nodes_by_pos[pos] = node
                self.nodes_by_id[node.id] = node
                id += 1
        for y in range(1, size):
            for x in range(y, size * 2 - 1):
                ry = size + y - 1
                pos = (x, ry)
                node = Node(pos, id, [])
                self.nodes_by_pos[pos] = node
                self.nodes_by_id[node.id] = node
                id += 1

    def link_nodes(self):
        for node in self.nodes_by_id:
            (x, y) = node.pos
            for dir in DIRS:
                nx = x + dir.x
                ny = y + dir.y
                if self.contains_pos((nx, ny)):
                    node.links.append(self.nodes_by_pos[(nx, ny)].id)

    def contains_pos(self, pos):
        return pos in self.nodes_by_pos

    def get_by_pos(self, pos):
        return self.nodes_by_pos[pos]

    def get_by_id(self, id):
        return self.nodes_by_id[id]


##################################
class Pos(object):

    def __init__(self, hex, tiles, done=None):
        self.hex = hex
        self.tiles = tiles
        self.done = Done(hex.count) if done is None else done

    def clone(self):
        return Pos(self.hex, self.tiles, self.done.clone())

##################################


def constraint_pass(pos, last_move=None):
    changed = False
    left = pos.tiles[:]
    done = pos.done

    # Remove impossible values from free cells
    free_cells = (range(done.count) if last_move is None
                  else pos.hex.get_by_id(last_move).links)
    for i in free_cells:
        if not done.already_done(i):
            vmax = 0
            vmin = 0
            cells_around = pos.hex.get_by_id(i).links
            for nid in cells_around:
                if done.already_done(nid):
                    if done[nid][0] != EMPTY:
                        vmin += 1
                        vmax += 1
                else:
                    vmax += 1

            for num in range(7):
                if (num < vmin) or (num > vmax):
                    if done.remove(i, num):
                        changed = True

    # Computes how many of each value is still free
    for cell in done.cells:
        if len(cell) == 1:
            left[cell[0]] -= 1

    for v in range(8):
        # If there is none, remove the possibility from all tiles
        if (pos.tiles[v] > 0) and (left[v] == 0):
            if done.remove_unfixed(v):
                changed = True
        else:
            possible = sum((1 if v in cell else 0) for cell in done.cells)
            # If the number of possible cells for a value is exactly the number of available tiles
            # put a tile in each cell
            if pos.tiles[v] == possible:
                for i in range(done.count):
                    cell = done.cells[i]
                    if (not done.already_done(i)) and (v in cell):
                        done.set_done(i, v)
                        changed = True

    # Force empty or non-empty around filled cells
    filled_cells = (range(done.count) if last_move is None
                    else [last_move])
    for i in filled_cells:
        if done.already_done(i):
            num = done[i][0]
            empties = 0
            filled = 0
            unknown = []
            cells_around = pos.hex.get_by_id(i).links
            for nid in cells_around:
                if done.already_done(nid):
                    if done[nid][0] == EMPTY:
                        empties += 1
                    else:
                        filled += 1
                else:
                    unknown.append(nid)
            if len(unknown) > 0:
                if num == filled:
                    for u in unknown:
                        if EMPTY in done[u]:
                            done.set_done(u, EMPTY)
                            changed = True
                        # else:
                        #    raise Exception("Houston, we've got a problem")
                elif num == filled + len(unknown):
                    for u in unknown:
                        if done.remove(u, EMPTY):
                            changed = True

    return changed


ASCENDING = 1
DESCENDING = -1


def find_moves(pos, strategy, order):
    done = pos.done
    cell_id = done.next_cell(pos, strategy)
    if cell_id < 0:
        return []

    if order == ASCENDING:
        return [(cell_id, v) for v in done[cell_id]]
    else:
        # Try higher values first and EMPTY last
        moves = list(reversed([(cell_id, v)
                               for v in done[cell_id] if v != EMPTY]))
        if EMPTY in done[cell_id]:
            moves.append((cell_id, EMPTY))
        return moves


def play_move(pos, move):
    (cell_id, i) = move
    pos.done.set_done(cell_id, i)


def print_pos(pos, output):
    hex = pos.hex
    done = pos.done
    size = hex.size
    for y in range(size):
        print(" " * (size - y - 1), end="", file=output)
        for x in range(size + y):
            pos2 = (x, y)
            id = hex.get_by_pos(pos2).id
            if done.already_done(id):
                c = str(done[id][0]) if done[id][0] != EMPTY else "."
            else:
                c = "?"
            print("%s " % c, end="", file=output)
        print(end="\n", file=output)
    for y in range(1, size):
        print(" " * y, end="", file=output)
        for x in range(y, size * 2 - 1):
            ry = size + y - 1
            pos2 = (x, ry)
            id = hex.get_by_pos(pos2).id
            if done.already_done(id):
                c = str(done[id][0]) if done[id][0] != EMPTY else "."
            else:
                c = "?"
            print("%s " % c, end="", file=output)
        print(end="\n", file=output)


OPEN = 0
SOLVED = 1
IMPOSSIBLE = -1


def solved(pos, output, verbose=False):
    hex = pos.hex
    tiles = pos.tiles[:]
    done = pos.done
    exact = True
    all_done = True
    for i in range(hex.count):
        if len(done[i]) == 0:
            return IMPOSSIBLE
        elif done.already_done(i):
            num = done[i][0]
            tiles[num] -= 1
            if (tiles[num] < 0):
                return IMPOSSIBLE
            vmax = 0
            vmin = 0
            if num != EMPTY:
                cells_around = hex.get_by_id(i).links
                for nid in cells_around:
                    if done.already_done(nid):
                        if done[nid][0] != EMPTY:
                            vmin += 1
                            vmax += 1
                    else:
                        vmax += 1

                if (num < vmin) or (num > vmax):
                    return IMPOSSIBLE
                if num != vmin:
                    exact = False
        else:
            all_done = False

    if (not all_done) or (not exact):
        return OPEN

    print_pos(pos, output)
    return SOLVED


def solve_step(prev, strategy, order, output, first=False):
    if first:
        pos = prev.clone()
        while constraint_pass(pos):
            pass
    else:
        pos = prev

    moves = find_moves(pos, strategy, order)
    if len(moves) == 0:
        return solved(pos, output)
    else:
        for move in moves:
            # print("Trying (%d, %d)" % (move[0], move[1]))
            ret = OPEN
            new_pos = pos.clone()
            play_move(new_pos, move)
            # print_pos(new_pos)
            while constraint_pass(new_pos, move[0]):
                pass
            cur_status = solved(new_pos, output)
            if cur_status != OPEN:
                ret = cur_status
            else:
                ret = solve_step(new_pos, strategy, order, output)
            if ret == SOLVED:
                return SOLVED
    return IMPOSSIBLE


def check_valid(pos):
    hex = pos.hex
    tiles = pos.tiles
    # fill missing entries in tiles
    tot = 0
    for i in range(8):
        if tiles[i] > 0:
            tot += tiles[i]
        else:
            tiles[i] = 0
    # check total
    if tot != hex.count:
        raise Exception(
            "Invalid input. Expected %d tiles, got %d." % (hex.count, tot))


def solve(pos, strategy, order, output):
    check_valid(pos)
    return solve_step(pos, strategy, order, output, first=True)


# TODO Write an 'iterator' to go over all x,y positions

def read_file(file):
    lines = [line.strip("\r\n") for line in file.splitlines()]
    size = int(lines[0])
    hex = Hex(size)
    linei = 1
    tiles = 8 * [0]
    done = Done(hex.count)
    for y in range(size):
        line = lines[linei][size - y - 1:]
        p = 0
        for x in range(size + y):
            tile = line[p:p + 2]
            p += 2
            if tile[1] == ".":
                inctile = EMPTY
            else:
                inctile = int(tile)
            tiles[inctile] += 1
            # Look for locked tiles
            if tile[0] == "+":
                # print("Adding locked tile: %d at pos %d, %d, id=%d" %
                #      (inctile, x, y, hex.get_by_pos((x, y)).id))
                done.set_done(hex.get_by_pos((x, y)).id, inctile)

        linei += 1
    for y in range(1, size):
        ry = size - 1 + y
        line = lines[linei][y:]
        p = 0
        for x in range(y, size * 2 - 1):
            tile = line[p:p + 2]
            p += 2
            if tile[1] == ".":
                inctile = EMPTY
            else:
                inctile = int(tile)
            tiles[inctile] += 1
            # Look for locked tiles
            if tile[0] == "+":
                # print("Adding locked tile: %d at pos %d, %d, id=%d" %
                #      (inctile, x, ry, hex.get_by_pos((x, ry)).id))
                done.set_done(hex.get_by_pos((x, ry)).id, inctile)
        linei += 1
    hex.link_nodes()
    done.filter_tiles(tiles)
    return Pos(hex, tiles, done)


def solve_file(file, strategy, order, output):
    pos = read_file(file)
    solve(pos, strategy, order, output)


LEVELS = {}

LEVELS[2] = ("""
2
  . 1
 . 1 1
  1 .
""", """\
 1 1
. . .
 1 1
""")

LEVELS[10] = ("""
3
  +.+. .
 +. 0 . 2
 . 1+2 1 .
  2 . 0+.
   .+.+.
""", """\
  . . 1
 . 1 . 2
0 . 2 2 .
 . . . .
  0 . .
""")

LEVELS[20] = ("""
3
   . 5 4
  . 2+.+1
 . 3+2 3 .
 +2+. 5 .
   . 3 .
""", """\
  3 3 2
 4 5 . 1
3 5 2 . .
 2 . . .
  . . .
""")

LEVELS[25] = ("""
3
   4 . .
  . . 2 .
 4 3 2 . 4
  2 2 3 .
   4 2 4
""", """\
  3 4 2
 2 4 4 .
. . . 4 2
 . 2 4 3
  . 2 .
""")

LEVELS[30] = ("""
4
    5 5 . .
   3 . 2+2 6
  3 . 2 . 5 .
 . 3 3+4 4 . 3
  4 5 4 . 5 4
   5+2 . . 3
    4 . . .
""", """\
   3 4 3 .
  4 6 5 2 .
 2 5 5 . . 2
. . 5 4 . 4 3
 . 3 5 4 5 4
  . 2 . 3 3
   . . . .
""")

LEVELS[36] = ("""
4
    2 1 1 2
   3 3 3 . .
  2 3 3 . 4 .
 . 2 . 2 4 3 2
  2 2 . . . 2
   4 3 4 . .
    3 2 3 3
""", """\
   3 4 3 2
  3 4 4 . 3
 2 . . 3 4 3
2 . 1 . 3 . 2
 3 3 . 2 . 2
  3 . 2 . 2
   2 2 . 1
""")


def main(loops, level):
    board, solution = LEVELS[level]
    order = DESCENDING
    strategy = Done.FIRST_STRATEGY
    stream = io.StringIO()

    board = board.strip()
    expected = solution.rstrip()

    range_it = range(loops)
    t0 = pyperf.perf_counter()

    for _ in range_it:
        stream = io.StringIO()
        solve_file(board, strategy, order, stream)
        output = stream.getvalue()
        stream = None

    dt = pyperf.perf_counter() - t0

    output = '\n'.join(line.rstrip() for line in output.splitlines())
    if output != expected:
        raise AssertionError("got a wrong answer:\n%s\nexpected: %s"
                             % (output, expected))

    return dt


def add_cmdline_args(cmd, args):
    cmd.extend(("--level", str(args.level)))


if __name__ == "__main__":
    runner = pyperf.Runner(add_cmdline_args=add_cmdline_args)
    levels = sorted(LEVELS)
    runner.argparser.add_argument("--level", type=int,
                                  choices=levels,
                                  default=DEFAULT_LEVEL,
                                  help="Hexiom board level (default: %s)"
                                       % DEFAULT_LEVEL)

    args = runner.parse_args()
    runner.metadata['description'] = "Solver of Hexiom board game"
    runner.metadata['hexiom_level'] = args.level

    runner.bench_time_func('hexiom', main, args.level)



import pyperf
import gc

CYCLES = 100
LINKS = 20


class Node:
    def __init__(self):
        self.next = None
        self.prev = None

    def link_next(self, next):
        self.next = next
        self.next.prev = self


def create_cycle(node, n_links):
    """Create a cycle of n_links nodes, starting with node."""

    if n_links == 0:
        return

    current = node
    for i in range(n_links):
        next_node = Node()
        current.link_next(next_node)
        current = next_node

    current.link_next(node)


def create_gc_cycles(n_cycles, n_links):
    """Create n_cycles cycles n_links+1 nodes each."""

    cycles = []
    for _ in range(n_cycles):
        node = Node()
        cycles.append(node)
        create_cycle(node, n_links)
    return cycles


def benchamark_collection(loops, cycles, links):
    total_time = 0
    for _ in range(loops):
        gc.collect()
        all_cycles = create_gc_cycles(cycles, links)

        # Main loop to measure
        del all_cycles
        t0 = pyperf.perf_counter()
        collected = gc.collect()
        total_time += pyperf.perf_counter() - t0

        assert collected is None or collected >= cycles * (links + 1)

    return total_time


if __name__ == "__main__":
    runner = pyperf.Runner()
    runner.metadata["description"] = "GC link benchmark"
    runner.bench_time_func("create_gc_cycles", benchamark_collection, CYCLES, LINKS)



"""
Test the performance of simple HTTP serving with FastAPI.

This benchmark tests FastAPI's request handling, including:
- Path parameter extraction and validation
- Pydantic model serialization
- JSON response encoding

The bench serves a REST API endpoint that returns JSON objects,
simulating a typical web application scenario.

Author: Savannah Ostrowski
"""

import asyncio
import socket

import httpx
import pyperf
import threading
import uvicorn
from fastapi import FastAPI
from pydantic import BaseModel

HOST = "127.0.0.1"

CONCURRENCY = 150

class Item(BaseModel):
    id: int
    name: str
    price: float
    tags: list[str] = []

app = FastAPI()

@app.get("/items/{item_id}", response_model=Item)
async def get_item(item_id: int):
    return {
        "id": item_id,
        "name": "Sample Item",
        "price": 9.99,
        "tags": ["sample", "item", "fastapi"]
    }

def setup_server(): 
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind((HOST, 0))
        s.listen(1)
        port = s.getsockname()[1]

    config = uvicorn.Config(app, host=HOST, port=port, log_level="error")
    server = uvicorn.Server(config)

    server_thread = threading.Thread(target=server.run, daemon=True)
    server_thread.start()

    while not server.started:
        pass

    url = f"http://{HOST}:{port}"
    return url

def bench_fastapi(loops, url):
    async def run_benchmark():
        async with httpx.AsyncClient() as client:
            t0 = pyperf.perf_counter()

            for i in range(loops):
                tasks = [
                    client.get(f"{url}/items/{i}")
                    for _ in range(CONCURRENCY)
                ]
                responses = await asyncio.gather(*tasks)
                for response in responses:
                    response.raise_for_status()
                    data = response.json()
                    assert data["id"] == i
                    assert "tags" in data

            return pyperf.perf_counter() - t0

    return asyncio.run(run_benchmark())


if __name__ == "__main__":
    url = setup_server()
    runner = pyperf.Runner()
    runner.metadata['description'] = "Test the performance of HTTP requests with FastAPI"
    runner.bench_time_func("fastapi_http", bench_fastapi, url)



"""Wrapper script for testing the performance of the html5lib HTML 5 parser.

The input data is the spec document for HTML 5, written in HTML 5.
The spec was pulled from http://svn.whatwg.org/webapps/index.
"""
import io
import os.path

import html5lib
import pyperf


__author__ = "collinwinter@google.com (Collin Winter)"


def bench_html5lib(html_file):
    html_file.seek(0)
    html5lib.parse(html_file)


if __name__ == "__main__":
    runner = pyperf.Runner()
    runner.metadata['description'] = (
        "Test the performance of the html5lib parser.")
    runner.metadata['html5lib_version'] = html5lib.__version__

    # Get all our IO over with early.
    filename = os.path.join(os.path.dirname(__file__),
                            "data", "w3_tr_html5.html")
    with open(filename, "rb") as fp:
        html_file = io.BytesIO(fp.read())

    runner.bench_func('html5lib', bench_html5lib, html_file)



"""
Calculate `factorial` using the decimal module.

- 2024-06-14: Michael Droettboom copied this from
  Modules/_decimal/tests/bench.py in the CPython source and adapted to use
  pyperf.
"""

# Original copyright notice in CPython source:

#
# Copyright (C) 2001-2012 Python Software Foundation. All Rights Reserved.
# Modified and extended by Stefan Krah.
#


import decimal


import pyperf


def factorial(n, m):
    if n > m:
        return factorial(m, n)
    elif m == 0:
        return 1
    elif n == m:
        return n
    else:
        return factorial(n, (n + m) // 2) * factorial((n + m) // 2 + 1, m)


def bench_decimal_factorial():
    c = decimal.getcontext()
    c.prec = decimal.MAX_PREC
    c.Emax = decimal.MAX_EMAX
    c.Emin = decimal.MIN_EMIN

    for n in [10000, 100000]:
        # C version of decimal
        _ = factorial(decimal.Decimal(n), 0)


if __name__ == "__main__":
    runner = pyperf.Runner()
    runner.metadata["description"] = "decimal_factorial benchmark"

    args = runner.parse_args()
    runner.bench_func("decimal_factorial", bench_decimal_factorial)



"""
Script for testing the performance of logging simple messages.

Rationale for logging_silent by Antoine Pitrou:

"The performance of silent logging calls is actually important for all
applications which have debug() calls in their critical paths.  This is
quite common in network and/or distributed programming where you want to
allow logging many events for diagnosis of unexpected runtime issues
(because many unexpected conditions can appear), but with those logs
disabled by default for performance and readability reasons."

https://mail.python.org/pipermail/speed/2017-May/000576.html
"""

# Python imports
import io
import logging

# Third party imports
import pyperf

# A simple format for parametered logging
FORMAT = 'important: %s'
MESSAGE = 'some important information to be logged'


def truncate_stream(stream):
    stream.seek(0)
    stream.truncate()


def bench_silent(loops, logger, stream):
    truncate_stream(stream)

    # micro-optimization: use fast local variables
    m = MESSAGE
    range_it = range(loops)
    t0 = pyperf.perf_counter()

    for _ in range_it:
        # repeat 10 times
        logger.debug(m)
        logger.debug(m)
        logger.debug(m)
        logger.debug(m)
        logger.debug(m)
        logger.debug(m)
        logger.debug(m)
        logger.debug(m)
        logger.debug(m)
        logger.debug(m)

    dt = pyperf.perf_counter() - t0

    if len(stream.getvalue()) != 0:
        raise ValueError("stream is expected to be empty")

    return dt


def bench_simple_output(loops, logger, stream):
    truncate_stream(stream)

    # micro-optimization: use fast local variables
    m = MESSAGE
    range_it = range(loops)
    t0 = pyperf.perf_counter()

    for _ in range_it:
        # repeat 10 times
        logger.warning(m)
        logger.warning(m)
        logger.warning(m)
        logger.warning(m)
        logger.warning(m)
        logger.warning(m)
        logger.warning(m)
        logger.warning(m)
        logger.warning(m)
        logger.warning(m)

    dt = pyperf.perf_counter() - t0

    lines = stream.getvalue().splitlines()
    if len(lines) != loops * 10:
        raise ValueError("wrong number of lines")

    return dt


def bench_formatted_output(loops, logger, stream):
    truncate_stream(stream)

    # micro-optimization: use fast local variables
    fmt = FORMAT
    msg = MESSAGE
    range_it = range(loops)
    t0 = pyperf.perf_counter()

    for _ in range_it:
        # repeat 10 times
        logger.warning(fmt, msg)
        logger.warning(fmt, msg)
        logger.warning(fmt, msg)
        logger.warning(fmt, msg)
        logger.warning(fmt, msg)
        logger.warning(fmt, msg)
        logger.warning(fmt, msg)
        logger.warning(fmt, msg)
        logger.warning(fmt, msg)
        logger.warning(fmt, msg)

    dt = pyperf.perf_counter() - t0

    lines = stream.getvalue().splitlines()
    if len(lines) != loops * 10:
        raise ValueError("wrong number of lines")

    return dt


def add_cmdline_args(cmd, args):
    if args.benchmark:
        cmd.append(args.benchmark)


BENCHMARKS = {
    "silent": bench_silent,
    "simple": bench_simple_output,
    "format": bench_formatted_output,
}


if __name__ == "__main__":
    runner = pyperf.Runner(add_cmdline_args=add_cmdline_args)
    runner.metadata['description'] = "Test the performance of logging."

    parser = runner.argparser
    parser.add_argument("benchmark", nargs='?', choices=sorted(BENCHMARKS))

    options = runner.parse_args()

    # NOTE: StringIO performance will impact the results...
    stream = io.StringIO()

    handler = logging.StreamHandler(stream=stream)
    logger = logging.getLogger("benchlogger")
    logger.propagate = False
    logger.addHandler(handler)
    logger.setLevel(logging.WARNING)

    if options.benchmark:
        benchmarks = (options.benchmark,)
    else:
        benchmarks = sorted(BENCHMARKS)

    for bench in benchmarks:
        name = 'logging_%s' % bench
        bench_func = BENCHMARKS[bench]

        runner.bench_time_func(name, bench_func,
                               logger, stream,
                               inner_loops=10)



"""
Benchmark for concurrent model communication.
"""
import pyperf

from multiprocessing.pool import Pool, ThreadPool


def f(x: int) -> int:
    return x


def bench_mp_pool(p: int, n: int, chunk: int) -> None:
    with Pool(p) as pool:
        for _ in pool.imap(f, range(n), chunk):
            pass


def bench_thread_pool(c: int, n: int, chunk: int) -> None:
    with ThreadPool(c) as pool:
        for _ in pool.imap(f, range(n), chunk):
            pass


if __name__ == "__main__":
    runner = pyperf.Runner()
    runner.metadata["description"] = "concurrent model communication benchmark"
    count = 1000
    chunk = 10
    num_core = 2
    runner.bench_func("bench_mp_pool", bench_mp_pool, num_core, count, chunk)
    runner.bench_func("bench_thread_pool", bench_thread_pool, num_core, count, chunk)



"""
Benchmark for asyncio TCP server and client performance
transferring 10MB of data.

Author: Kumar Aditya
"""


import asyncio
from pyperf import Runner
import ssl
import os

CHUNK_SIZE = 1024 ** 2 * 10
# Taken from CPython's test suite
SSL_CERT = os.path.join(os.path.dirname(__file__), 'ssl_cert.pem')
SSL_KEY = os.path.join(os.path.dirname(__file__), 'ssl_key.pem')


async def handle_echo(reader: asyncio.StreamReader,
                      writer: asyncio.StreamWriter) -> None:
    data = b'x' * CHUNK_SIZE
    for _ in range(100):
        writer.write(data)
        await writer.drain()
    writer.close()
    await writer.wait_closed()


async def main(use_ssl: bool) -> None:
    if use_ssl:
        server_context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
        server_context.load_cert_chain(SSL_CERT, SSL_KEY)
        server_context.check_hostname = False
        server_context.verify_mode = ssl.CERT_NONE

        client_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        client_context.load_cert_chain(SSL_CERT, SSL_KEY)
        client_context.check_hostname = False
        client_context.verify_mode = ssl.CERT_NONE
    else:
        server_context = None
        client_context = None

    server = await asyncio.start_server(handle_echo, '127.0.0.1', 8882, ssl=server_context)

    async with server:
        asyncio.create_task(server.start_serving())
        reader, writer = await asyncio.open_connection('127.0.0.1', 8882, ssl=client_context)
        data_len = 0
        while True:
            data = await reader.read(CHUNK_SIZE)
            if not data:
                break
            data_len += len(data)
        assert data_len == CHUNK_SIZE * 100
        writer.close()
        await writer.wait_closed()


def add_cmdline_args(cmd, args):
    if args.ssl:
        cmd.append("--ssl")


if __name__ == '__main__':
    runner = Runner(add_cmdline_args=add_cmdline_args)
    parser = runner.argparser
    parser.add_argument('--ssl', action='store_true', default=False)
    args = runner.parse_args()
    name = 'asyncio_tcp' + ('_ssl' if args.ssl else '')
    runner.bench_async_func(name, main, args.ssl)



import pathlib
import tempfile
import unittest

from pyperformance import _pyproject_toml


class TestPyProjectToml(unittest.TestCase):
    def test_parse_project_with_readme_string(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            tmp_path = pathlib.Path(tmpdir)
            toml_content = """
            [project]
            name = "my-test-bench"
            version = "1.0"
            readme = "README.md"
            """
            (tmp_path / "README.md").touch()

            data = _pyproject_toml.parse_pyproject_toml(
                toml_content, rootdir=str(tmp_path), requirefiles=True
            )
            self.assertEqual(data["project"]["readme"]["file"], "README.md")

    def test_parse_full_valid_project_section(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            toml_content = """
            [project]
            name = "my-full-bench"
            version = "2.1.3"
            dependencies = [
                "pyperf",
                "six",
            ]
            requires-python = ">=3.12"
            """
            data = _pyproject_toml.parse_pyproject_toml(
                toml_content, rootdir=str(tmpdir)
            )
            project = data["project"]
            self.assertEqual(project["name"], "my-full-bench")
            self.assertEqual(project["version"], "2.1.3")
            self.assertEqual(project["dependencies"], ["pyperf", "six"])
            self.assertEqual(project["requires-python"], ">=3.12")

    def test_parse_fails_on_missing_name(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            toml_content = """
            [project]
            version = "1.0"
            """

            with self.assertRaisesRegex(ValueError, r'missing required "name" field'):
                _pyproject_toml.parse_pyproject_toml(toml_content, rootdir=str(tmpdir))

    def test_parse_fails_on_unsupported_section(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            toml_content = """
            [project]
            name = "my-test-bench"
            version = "1.0"

            [foobar]
            key = "value"
            """

            with self.assertRaisesRegex(ValueError, "unsupported sections"):
                _pyproject_toml.parse_pyproject_toml(toml_content, rootdir=str(tmpdir))

    def test_parse_readme_file_missing_with_requirefiles_true(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            toml_content = """
            [project]
            name = "my-test-bench"
            version = "1.0"
            readme = "MISSING_README.md"
            """

            with self.assertRaisesRegex(ValueError, "does not exist"):
                _pyproject_toml.parse_pyproject_toml(
                    toml_content, rootdir=str(tmpdir), requirefiles=True
                )

    def test_parse_readme_file_missing_with_requirefiles_false(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            toml_content = """
            [project]
            name = "my-test-bench"
            version = "1.0"
            readme = "MISSING_README.md"
            """
            data = _pyproject_toml.parse_pyproject_toml(
                toml_content, rootdir=str(tmpdir), requirefiles=False
            )
            self.assertEqual(data["project"]["readme"]["file"], "MISSING_README.md")


if __name__ == "__main__":
    unittest.main()




"""Benchmark script for testing the performance of ElementTree.

This is intended to support Unladen Swallow's pyperf.py.

This will have ElementTree, cElementTree and lxml (if available)
parse a generated XML file, search it, create new XML trees from
it and serialise the result.
"""

import io
import os
import sys
import tempfile
from collections import defaultdict

import pyperf


__author__ = "stefan_ml@behnel.de (Stefan Behnel)"

FALLBACK_ETMODULE = 'xml.etree.ElementTree'


def build_xml_tree(etree):
    SubElement = etree.SubElement
    root = etree.Element('root')

    # create a couple of repetitive broad subtrees
    for c in range(50):
        child = SubElement(root, 'child-%d' % c,
                                 tag_type="child")
        for i in range(100):
            SubElement(child, 'subchild').text = 'LEAF-%d-%d' % (c, i)

    # create a deep subtree
    deep = SubElement(root, 'deepchildren', tag_type="deepchild")
    for i in range(50):
        deep = SubElement(deep, 'deepchild')
    SubElement(deep, 'deepleaf', tag_type="leaf").text = "LEAF"

    # store the number of elements for later
    nb_elems = sum(1 for elem in root.iter())
    root.set('nb-elems', str(nb_elems))

    return root


def process(etree, xml_root=None):
    SubElement = etree.SubElement

    if xml_root is not None:
        root = xml_root
    else:
        root = build_xml_tree(etree)

    # find*()
    found = sum(child.find('.//deepleaf') is not None
                for child in root)
    if found != 1:
        raise RuntimeError("find() failed")

    text = 'LEAF-5-99'
    found = any(1 for child in root
                for el in child.iterfind('.//subchild')
                if el.text == text)
    if not found:
        raise RuntimeError("iterfind() failed")

    found = sum(el.text == 'LEAF'
                for el in root.findall('.//deepchild/deepleaf'))
    if found != 1:
        raise RuntimeError("findall() failed")

    # tree creation based on original tree
    dest = etree.Element('root2')
    target = SubElement(dest, 'result-1')
    for child in root:
        SubElement(target, child.tag).text = str(len(child))
    if len(target) != len(root):
        raise RuntimeError("transform #1 failed")

    target = SubElement(dest, 'result-2')
    for child in root.iterfind('.//subchild'):
        SubElement(target, child.tag, attr=child.text).text = "found"

    if (len(target) < len(root)
            or not all(el.text == 'found'
                       for el in target.iterfind('subchild'))):
        raise RuntimeError("transform #2 failed")

    # moving subtrees around
    orig_len = len(root[0])
    new_root = root.makeelement('parent', {})
    new_root[:] = root[0]
    el = root[0]
    del el[:]
    for child in new_root:
        if child is not None:
            el.append(child)
    if len(el) != orig_len:
        raise RuntimeError("child moving failed")

    # check iteration tree consistency
    d = defaultdict(list)
    for child in root:
        tags = d[child.get('tag_type')]
        for sub in child.iter():
            tags.append(sub)

    check_dict = dict((n, iter(ch)) for n, ch in d.items())
    target = SubElement(dest, 'transform-2')
    for child in root:
        tags = check_dict[child.get('tag_type')]
        for sub in child.iter():
            # note: explicit object identity check to make sure
            # users can properly keep state in the tree
            if sub is not next(tags):
                raise RuntimeError("tree iteration consistency check failed")
            SubElement(target, sub.tag).text = 'worked'

    # final probability check for serialisation (we added enough content
    # to make the result tree larger than the original)
    orig = etree.tostring(root, encoding='utf8')
    result = etree.tostring(dest, encoding='utf8')
    if (len(result) < len(orig)
            or b'worked' not in result
            or b'>LEAF<' not in orig):
        raise RuntimeError("serialisation probability check failed")
    return result


def bench_iterparse(etree, xml_file, xml_data, xml_root):
    for _ in range(10):
        it = etree.iterparse(xml_file, ('start', 'end'))
        events1 = [(event, elem.tag) for event, elem in it]
        it = etree.iterparse(io.BytesIO(xml_data), ('start', 'end'))
        events2 = [(event, elem.tag) for event, elem in it]
    nb_elems = int(xml_root.get('nb-elems'))
    if len(events1) != 2 * nb_elems or events1 != events2:
        raise RuntimeError("parsing check failed:\n%r\n%r\n" %
                           (len(events1), events2[:10]))


def bench_parse(etree, xml_file, xml_data, xml_root):
    for _ in range(30):
        root1 = etree.parse(xml_file).getroot()
        root2 = etree.fromstring(xml_data)
    result1 = etree.tostring(root1)
    result2 = etree.tostring(root2)
    if result1 != result2:
        raise RuntimeError("serialisation check failed")


def bench_process(etree, xml_file, xml_data, xml_root):
    result1 = process(etree, xml_root=xml_root)
    result2 = process(etree, xml_root=xml_root)
    if result1 != result2 or b'>found<' not in result2:
        raise RuntimeError("serialisation check failed")


def bench_generate(etree, xml_file, xml_data, xml_root):
    output = []
    for _ in range(10):
        root = build_xml_tree(etree)
        output.append(etree.tostring(root))

    length = None
    for xml in output:
        if length is None:
            length = len(xml)
        elif length != len(xml):
            raise RuntimeError("inconsistent output detected")
        if b'>LEAF<' not in xml:
            raise RuntimeError("unexpected output detected")


def bench_etree(iterations, etree, bench_func):
    xml_root = build_xml_tree(etree)
    xml_data = etree.tostring(xml_root)

    # not using NamedTemporaryFile() here as re-opening it is not portable
    tf, file_path = tempfile.mkstemp()
    try:
        etree.ElementTree(xml_root).write(file_path)

        t0 = pyperf.perf_counter()

        for _ in range(iterations):
            bench_func(etree, file_path, xml_data, xml_root)

        dt = pyperf.perf_counter() - t0
    finally:
        try:
            os.close(tf)
        except EnvironmentError:
            pass
        try:
            os.unlink(file_path)
        except EnvironmentError:
            pass

    return dt


BENCHMARKS = 'parse iterparse generate process'.split()


def add_cmdline_args(cmd, args):
    cmd.extend(("--etree-module", args.etree_module))
    if args.no_accelerator:
        cmd.append("--no-accelerator")
    if args.benchmark:
        cmd.append(args.benchmark)


if __name__ == "__main__":
    # On Python 3, xml.etree.cElementTree is a deprecated alias
    # to xml.etree.ElementTree
    default_etmodule = "xml.etree.ElementTree"

    runner = pyperf.Runner(add_cmdline_args=add_cmdline_args)
    runner.metadata['description'] = ("Test the performance of "
                                      "ElementTree XML processing.")

    parser = runner.argparser
    parser.add_argument("--etree-module", default=None, metavar="FQMN",
                        help="Select an ElementTree module to use "
                             "(fully qualified module name). "
                             "Default is '%s'" % default_etmodule)
    parser.add_argument("--no-accelerator", action="store_true", default=False,
                        help="Disable the '_elementree' accelerator module "
                             "for ElementTree.")
    parser.add_argument("benchmark", nargs='?', choices=BENCHMARKS)

    options = runner.parse_args()

    if not options.etree_module:
        if options.no_accelerator:
            options.etree_module = FALLBACK_ETMODULE
        else:
            options.etree_module = default_etmodule
    if options.no_accelerator:
        # prevent C accelerator from being used in 3.3
        sys.modules['_elementtree'] = None
        import xml.etree.ElementTree as et
        if et.SubElement.__module__ != 'xml.etree.ElementTree':
            raise RuntimeError("Unexpected C accelerator for ElementTree")

    try:
        from importlib import import_module
    except ImportError:
        def import_module(module_name):
            __import__(module_name)
            return sys.modules[module_name]

    try:
        etree_module = import_module(options.etree_module)
    except ImportError:
        if options.etree_module != default_etmodule:
            raise
        etree_module = import_module(FALLBACK_ETMODULE)

    # Fill elementtree_module metadata: check if the accelerator is used
    module = etree_module.__name__
    # xml.etree.ElementTree._Element_Py was added to Python 3.4
    if hasattr(etree_module, '_Element_Py'):
        accelerator = (etree_module.Element is not etree_module._Element_Py)
    else:
        if options.no_accelerator:
            accelerator = False
        else:
            accelerator = True
    if accelerator:
        module += ' (with C accelerator)'
    else:
        module += ' (pure Python)'
    runner.metadata['elementtree_module'] = module

    if options.benchmark:
        benchmarks = (options.benchmark,)
    else:
        benchmarks = BENCHMARKS

    # Run the benchmark
    for bench in benchmarks:
        if accelerator:
            name = 'xml_etree_%s' % bench
        else:
            name = 'xml_etree_pure_python_%s' % bench
        bench_func = globals()['bench_%s' % bench]
        runner.bench_time_func(name, bench_etree, etree_module, bench_func)



"""
Benchmark for async tree workload, which calls asyncio.gather() on a tree
(6 levels deep, 6 branches per level) with the leaf nodes simulating some
(potentially) async work (depending on the benchmark variant). Benchmark
variants include:

1) "none": No actual async work in the async tree.
2) "io": All leaf nodes simulate async IO workload (async sleep 50ms).
3) "memoization": All leaf nodes simulate async IO workload with 90% of
                  the data memoized
4) "cpu_io_mixed": Half of the leaf nodes simulate CPU-bound workload and
                   the other half simulate the same workload as the
                   "memoization" variant.

All variants also have an "eager" flavor that uses the asyncio eager task
factory (if available), and a "tg" variant that uses TaskGroups.
"""


import asyncio
import math
import random

import pyperf


NUM_RECURSE_LEVELS = 6
NUM_RECURSE_BRANCHES = 6
RANDOM_SEED = 0
IO_SLEEP_TIME = 0.05
MEMOIZABLE_PERCENTAGE = 90
CPU_PROBABILITY = 0.5
FACTORIAL_N = 500


class AsyncTree:
    def __init__(self, use_task_groups=False):
        self.cache = {}
        self.use_task_groups = use_task_groups
        # set to deterministic random, so that the results are reproducible
        random.seed(RANDOM_SEED)

    async def mock_io_call(self):
        await asyncio.sleep(IO_SLEEP_TIME)

    async def workload_func(self):
        raise NotImplementedError(
            "To be implemented by each variant's derived class."
        )

    async def recurse_with_gather(self, recurse_level):
        if recurse_level == 0:
            await self.workload_func()
            return

        await asyncio.gather(
            *[self.recurse_with_gather(recurse_level - 1)
              for _ in range(NUM_RECURSE_BRANCHES)]
        )

    async def recurse_with_task_group(self, recurse_level):
        if recurse_level == 0:
            await self.workload_func()
            return

        async with asyncio.TaskGroup() as tg:
            for _ in range(NUM_RECURSE_BRANCHES):
                tg.create_task(
                    self.recurse_with_task_group(recurse_level - 1))

    async def run(self):
        if self.use_task_groups:
            await self.recurse_with_task_group(NUM_RECURSE_LEVELS)
        else:
            await self.recurse_with_gather(NUM_RECURSE_LEVELS)


class EagerMixin:
    async def run(self):
        loop = asyncio.get_running_loop()
        if hasattr(asyncio, 'eager_task_factory'):
            loop.set_task_factory(asyncio.eager_task_factory)
        return await super().run()


class NoneAsyncTree(AsyncTree):
    async def workload_func(self):
        return


class EagerAsyncTree(EagerMixin, NoneAsyncTree):
    pass


class IOAsyncTree(AsyncTree):
    async def workload_func(self):
        await self.mock_io_call()


class EagerIOAsyncTree(EagerMixin, IOAsyncTree):
    pass


class MemoizationAsyncTree(AsyncTree):
    async def workload_func(self):
        # deterministic random, seed set in AsyncTree.__init__()
        data = random.randint(1, 100)

        if data <= MEMOIZABLE_PERCENTAGE:
            if self.cache.get(data):
                return data

            self.cache[data] = True

        await self.mock_io_call()
        return data


class EagerMemoizationAsyncTree(EagerMixin, MemoizationAsyncTree):
    pass


class CpuIoMixedAsyncTree(MemoizationAsyncTree):
    async def workload_func(self):
        # deterministic random, seed set in AsyncTree.__init__()
        if random.random() < CPU_PROBABILITY:
            # mock cpu-bound call
            return math.factorial(FACTORIAL_N)
        else:
            return await MemoizationAsyncTree.workload_func(self)


class EagerCpuIoMixedAsyncTree(EagerMixin, CpuIoMixedAsyncTree):
    pass


def add_metadata(runner):
    runner.metadata["description"] = "Async tree workloads."
    runner.metadata["async_tree_recurse_levels"] = NUM_RECURSE_LEVELS
    runner.metadata["async_tree_recurse_branches"] = NUM_RECURSE_BRANCHES
    runner.metadata["async_tree_random_seed"] = RANDOM_SEED
    runner.metadata["async_tree_io_sleep_time"] = IO_SLEEP_TIME
    runner.metadata["async_tree_memoizable_percentage"] = MEMOIZABLE_PERCENTAGE
    runner.metadata["async_tree_cpu_probability"] = CPU_PROBABILITY
    runner.metadata["async_tree_factorial_n"] = FACTORIAL_N


def add_cmdline_args(cmd, args):
    cmd.append(args.benchmark)
    if args.task_groups:
        cmd.append("--task-groups")


def add_parser_args(parser):
    parser.add_argument(
        "benchmark",
        choices=BENCHMARKS,
        help="""\
Determines which benchmark to run. Options:
1) "none": No actual async work in the async tree.
2) "io": All leaf nodes simulate async IO workload (async sleep 50ms).
3) "memoization": All leaf nodes simulate async IO workload with 90%% of
                  the data memoized
4) "cpu_io_mixed": Half of the leaf nodes simulate CPU-bound workload and
                   the other half simulate the same workload as the
                   "memoization" variant.
""",
    )
    parser.add_argument(
        "--task-groups",
        action="store_true",
        default=False,
        help="Use TaskGroups instead of gather.",
    )


BENCHMARKS = {
    "none": NoneAsyncTree,
    "eager": EagerAsyncTree,
    "io": IOAsyncTree,
    "eager_io": EagerIOAsyncTree,
    "memoization": MemoizationAsyncTree,
    "eager_memoization": EagerMemoizationAsyncTree,
    "cpu_io_mixed": CpuIoMixedAsyncTree,
    "eager_cpu_io_mixed": EagerCpuIoMixedAsyncTree,
}


if __name__ == "__main__":
    runner = pyperf.Runner(add_cmdline_args=add_cmdline_args)
    add_metadata(runner)
    add_parser_args(runner.argparser)
    args = runner.parse_args()
    benchmark = args.benchmark

    async_tree_class = BENCHMARKS[benchmark]
    async_tree = async_tree_class(use_task_groups=args.task_groups)
    bench_name = f"async_tree_{benchmark}"
    if args.task_groups:
        bench_name += "_tg"
    runner.bench_async_func(bench_name, async_tree.run)



import hashlib
import json
import os
import sys
import time
import traceback
from collections import namedtuple

import pyperformance

from . import _python, _pythoninfo, _utils, _venv
from .venv import REQUIREMENTS_FILE, VenvForBenchmarks


class BenchmarkException(Exception):
    pass


class RunID(namedtuple("RunID", "python compat bench timestamp")):
    def __new__(cls, python, compat, bench, timestamp):
        self = super().__new__(
            cls,
            python,
            compat,
            bench or None,
            int(timestamp) if timestamp else None,
        )
        return self

    def __str__(self):
        if not self.timestamp:
            return self.name
        return f"{self.name}-{self.timestamp}"

    @property
    def name(self):
        try:
            return self._name
        except AttributeError:
            name = f"{self.python}-compat-{self.compat}"
            if self.bench:
                name = f"{name}-bm-{self.bench.name}"
            self._name = name
            return self._name


def get_run_id(python, bench=None):
    py_id = _python.get_id(python, prefix=True)
    compat_id = get_compatibility_id(bench)
    ts = time.time()
    return RunID(py_id, compat_id, bench, ts)


def get_loops_from_file(filename):
    with open(filename) as fd:
        data = json.load(fd)

    loops = {}
    for benchmark in data["benchmarks"]:
        metadata = benchmark.get("metadata", data["metadata"])
        name = metadata["name"]
        if name.endswith("_none"):
            name = name[: -len("_none")]
        if "loops" in metadata:
            loops[name] = metadata["loops"]

    return loops


def run_benchmarks(should_run, python, options):
    if options.same_loops is not None:
        loops = get_loops_from_file(options.same_loops)
    else:
        loops = {}

    to_run = sorted(should_run)

    info = _pythoninfo.get_info(python)
    runid = get_run_id(info)

    unique = getattr(options, "unique_venvs", False)
    if not unique:
        common = VenvForBenchmarks.ensure(
            _venv.get_venv_root(runid.name, python=info),
            info,
            upgrade="oncreate",
            inherit_environ=options.inherit_environ,
        )

    benchmarks = {}
    venvs = set()
    for i, bench in enumerate(to_run):
        bench_runid = runid._replace(bench=bench)
        assert bench_runid.name, (bench, bench_runid)
        name = bench_runid.name
        venv_root = _venv.get_venv_root(name, python=info)
        print()
        print("=" * 50)
        print(f"({i + 1:>2}/{len(to_run)}) creating venv for benchmark ({bench.name})")
        print()
        if not unique:
            print("(trying common venv first)")
            # Try the common venv first.
            try:
                common.ensure_reqs(bench)
            except _venv.RequirementsInstallationFailedError:
                print("(falling back to unique venv)")
            else:
                benchmarks[bench] = (common, bench_runid)
                continue
        try:
            venv = VenvForBenchmarks.ensure(
                venv_root,
                info,
                upgrade="oncreate",
                inherit_environ=options.inherit_environ,
            )
            # XXX Do not override when there is a requirements collision.
            venv.ensure_reqs(bench)
        except _venv.RequirementsInstallationFailedError:
            print("(benchmark will be skipped)")
            print()
            venv = None
        venvs.add(venv_root)
        benchmarks[bench] = (venv, bench_runid)
    print()

    suite = None
    run_count = str(len(to_run))
    errors = []

    base_pyperf_opts = get_pyperf_opts(options)

    import pyperf

    for index, bench in enumerate(to_run):
        name = bench.name
        print("[%s/%s] %s..." % (str(index + 1).rjust(len(run_count)), run_count, name))
        sys.stdout.flush()

        def add_bench(dest_suite, obj):
            if isinstance(obj, pyperf.BenchmarkSuite):
                results = obj
            else:
                results = (obj,)

            version = pyperformance.__version__
            for res in results:
                res.update_metadata(
                    {"performance_version": version, "tags": bench.tags}
                )

                if dest_suite is not None:
                    dest_suite.add_benchmark(res)
                else:
                    dest_suite = pyperf.BenchmarkSuite([res])

            return dest_suite

        if name in loops:
            pyperf_opts = [*base_pyperf_opts, f"--loops={loops[name]}"]
        else:
            pyperf_opts = base_pyperf_opts

        bench_venv, bench_runid = benchmarks.get(bench)
        if bench_venv is None:
            print("ERROR: Benchmark %s failed: could not install requirements" % name)
            errors.append((name, "Install requirements error"))
            continue
        try:
            result = bench.run(
                bench_venv.python,
                bench_runid,
                pyperf_opts,
                venv=bench_venv,
                verbose=options.verbose,
            )
        except TimeoutError as exc:
            print("ERROR: Benchmark %s timed out" % name)
            errors.append((name, exc))
        except RuntimeError as exc:
            print("ERROR: Benchmark %s failed: %s" % (name, exc))
            traceback.print_exc()
            errors.append((name, exc))
        except Exception as exc:
            print("ERROR: Benchmark %s failed: %s" % (name, exc))
            traceback.print_exc()
            errors.append((name, exc))
        else:
            suite = add_bench(suite, result)

    print()

    return (suite, errors)


# Utility functions


def get_compatibility_id(bench=None):
    # XXX Do not include the pyperformance reqs if a benchmark was provided?
    reqs = sorted(_utils.iter_clean_lines(REQUIREMENTS_FILE))
    if bench:
        lockfile = bench.requirements_lockfile
        if lockfile and os.path.exists(lockfile):
            reqs += sorted(_utils.iter_clean_lines(lockfile))

    data = [
        # XXX Favor pyperf.__version__ instead?
        pyperformance.__version__,
        "\n".join(reqs),
    ]

    h = hashlib.sha256()
    for value in data:
        h.update(value.encode("utf-8"))
    compat_id = h.hexdigest()
    # XXX Return the whole string?
    compat_id = compat_id[:12]

    return compat_id


def get_pyperf_opts(options):
    opts = []

    if options.debug_single_value:
        opts.append("--debug-single-value")
    elif options.rigorous:
        opts.append("--rigorous")
    elif options.fast:
        opts.append("--fast")

    if options.verbose:
        opts.append("--verbose")

    if options.affinity:
        opts.append("--affinity=%s" % options.affinity)
    if options.track_memory:
        opts.append("--track-memory")
    if options.inherit_environ:
        opts.append("--inherit-environ=%s" % ",".join(options.inherit_environ))
    if options.min_time:
        opts.append("--min-time=%s" % options.min_time)
    if options.timeout:
        opts.append("--timeout=%s" % options.timeout)
    if options.hook:
        for hook in options.hook:
            opts.append("--hook=%s" % hook)
    # --warmups=0 is a valid option, so check for `not None` here
    if options.warmups is not None:
        opts.append("--warmups=%s" % options.warmups)

    return opts



# coding: UTF-8
""" Telco Benchmark for measuring the performance of decimal calculations

- http://speleotrove.com/decimal/telco.html
- http://speleotrove.com/decimal/telcoSpec.html

A call type indicator, c, is set from the bottom (least significant) bit of the duration (hence c is 0 or 1).
A rate, r, is determined from the call type. Those calls with c=0 have a low r: 0.0013; the remainder (‘distance calls’) have a ‘premium’ r: 0.00894. (The rates are, very roughly, in Euros or dollarates per second.)
A price, p, for the call is then calculated (p=r*n). This is rounded to exactly 2 fractional digits using round-half-even (Banker’s round to nearest).
A basic tax, b, is calculated: b=p*0.0675 (6.75%). This is truncated to exactly 2 fractional digits (round-down), and the total basic tax variable is then incremented (sumB=sumB+b).
For distance calls: a distance tax, d, is calculated: d=p*0.0341 (3.41%). This is truncated to exactly 2 fractional digits (round-down), and then the total distance tax variable is incremented (sumD=sumD+d).
The total price, t, is calculated (t=p+b, and, if a distance call, t=t+d).
The total prices variable is incremented (sumT=sumT+t).
The total price, t, is converted to a string, s.

"""

from decimal import ROUND_HALF_EVEN, ROUND_DOWN, Decimal, getcontext, Context
import io
import os
from struct import unpack

import pyperf


def rel_path(*path):
    return os.path.join(os.path.dirname(__file__), *path)


def bench_telco(loops, filename):
    getcontext().rounding = ROUND_DOWN
    rates = list(map(Decimal, ('0.0013', '0.00894')))
    twodig = Decimal('0.01')
    Banker = Context(rounding=ROUND_HALF_EVEN)
    basictax = Decimal("0.0675")
    disttax = Decimal("0.0341")

    with open(filename, "rb") as infil:
        data = infil.read()

    infil = io.BytesIO(data)
    outfil = io.StringIO()

    start = pyperf.perf_counter()
    for _ in range(loops):
        infil.seek(0)

        sumT = Decimal("0")   # sum of total prices
        sumB = Decimal("0")   # sum of basic tax
        sumD = Decimal("0")   # sum of 'distance' tax

        for i in range(5000):
            datum = infil.read(8)
            if datum == '':
                break
            n, =  unpack('>Q', datum)

            calltype = n & 1
            r = rates[calltype]

            p = Banker.quantize(r * n, twodig)

            b = p * basictax
            b = b.quantize(twodig)
            sumB += b

            t = p + b

            if calltype:
                d = p * disttax
                d = d.quantize(twodig)
                sumD += d
                t += d

            sumT += t
            print(t, file=outfil)

        outfil.seek(0)
        outfil.truncate()

    return pyperf.perf_counter() - start


if __name__ == "__main__":
    runner = pyperf.Runner()
    runner.metadata['description'] = "Telco decimal benchmark"

    filename = rel_path("data", "telco-bench.b")
    runner.bench_time_func('telco', bench_telco, filename)



"""
Calculate `pi` using the decimal module.

The `pidigits` benchmark does a similar thing using regular (long) ints.

- 2024-06-14: Michael Droettboom copied this from
  Modules/_decimal/tests/bench.py in the CPython source and adapted to use
  pyperf.
"""

# Original copyright notice in CPython source:

#
# Copyright (C) 2001-2012 Python Software Foundation. All Rights Reserved.
# Modified and extended by Stefan Krah.
#


import decimal


import pyperf


def pi_decimal():
    """decimal"""
    D = decimal.Decimal
    lasts, t, s, n, na, d, da = D(0), D(3), D(3), D(1), D(0), D(0), D(24)
    while s != lasts:
        lasts = s
        n, na = n + na, na + 8
        d, da = d + da, da + 32
        t = (t * n) / d
        s += t
    return s


def bench_decimal_pi():
    for prec in [9, 19]:
        decimal.getcontext().prec = prec
        for _ in range(10000):
            _ = pi_decimal()


if __name__ == "__main__":
    runner = pyperf.Runner()
    runner.metadata["description"] = "decimal_pi benchmark"

    args = runner.parse_args()
    runner.bench_func("decimal_pi", bench_decimal_pi)



"""
Benchmark ``loads()`` function of the ``tomli`` module
on a large TOML file of GitHub's real world data generated by
the ``generate_data.py`` script.

It heavily exercises string operations such as concatenation,
subscripting and iteration.

Author: Kumar Aditya
"""

from pathlib import Path

import pyperf
import tomli

DATA_FILE = Path(__file__).parent / "data" / "tomli-bench-data.toml"

def bench_tomli_loads(loops: int) -> float:
    data = DATA_FILE.read_text('utf-8')
    range_it = range(loops)
    t0 = pyperf.perf_counter()
    for _ in range_it:
        tomli.loads(data)
    return pyperf.perf_counter() - t0

if __name__ == "__main__":
    runner = pyperf.Runner()
    runner.metadata['description'] = "Benchmark tomli.loads()"
    runner.bench_time_func('tomli_loads', bench_tomli_loads)



"""
Quadtree N-body benchmark using the pyperf framework.

This benchmark simulates gravitational interactions between particles using
a quadtree for spatial partitioning and the Barnes-Hut approximation algorithm.
No visualization, pure Python implementation without dependencies.
"""

import pyperf
import math

DEFAULT_ITERATIONS = 100
DEFAULT_PARTICLES = 200
DEFAULT_THETA = 0.5

# Constants
G = 6.67430e-11  # Gravitational constant
SOFTENING = 5.0  # Softening parameter to avoid singularities
TIME_STEP = 0.1  # Time step for simulation

class Point:
    def __init__(self, x, y):
        self.x = x
        self.y = y

class Particle:
    def __init__(self, x, y, mass=1.0):
        self.position = Point(x, y)
        self.velocity = Point(0, 0)
        self.acceleration = Point(0, 0)
        self.mass = mass
    
    def update(self, time_step):
        # Update velocity using current acceleration
        self.velocity.x += self.acceleration.x * time_step
        self.velocity.y += self.acceleration.y * time_step
        
        # Update position using updated velocity
        self.position.x += self.velocity.x * time_step
        self.position.y += self.velocity.y * time_step
        
        # Reset acceleration for next frame
        self.acceleration.x = 0
        self.acceleration.y = 0
    
    def apply_force(self, fx, fy):
        # F = ma -> a = F/m
        self.acceleration.x += fx / self.mass
        self.acceleration.y += fy / self.mass

class Rectangle:
    def __init__(self, x, y, w, h):
        self.x = x
        self.y = y
        self.w = w
        self.h = h
    
    def contains(self, point):
        return (
            point.x >= self.x - self.w and
            point.x < self.x + self.w and
            point.y >= self.y - self.h and
            point.y < self.y + self.h
        )
    
    def intersects(self, range_rect):
        return not (
            range_rect.x - range_rect.w > self.x + self.w or
            range_rect.x + range_rect.w < self.x - self.w or
            range_rect.y - range_rect.h > self.y + self.h or
            range_rect.y + range_rect.h < self.y - self.h
        )

class QuadTree:
    def __init__(self, boundary, capacity=4):
        self.boundary = boundary
        self.capacity = capacity
        self.particles = []
        self.divided = False
        self.center_of_mass = Point(0, 0)
        self.total_mass = 0
        self.northeast = None
        self.northwest = None
        self.southeast = None
        self.southwest = None
    
    def insert(self, particle):
        if not self.boundary.contains(particle.position):
            return False
        
        if len(self.particles) < self.capacity and not self.divided:
            self.particles.append(particle)
            self.update_mass_distribution(particle)
            return True
        
        if not self.divided:
            self.subdivide()
        
        if self.northeast.insert(particle): return True
        if self.northwest.insert(particle): return True
        if self.southeast.insert(particle): return True
        if self.southwest.insert(particle): return True
        
        # This should never happen if the boundary check is correct
        return False
    
    def update_mass_distribution(self, particle):
        # Update center of mass and total mass when adding a particle
        total_mass_new = self.total_mass + particle.mass
        
        # Calculate new center of mass
        if total_mass_new > 0:
            self.center_of_mass.x = (self.center_of_mass.x * self.total_mass + 
                                    particle.position.x * particle.mass) / total_mass_new
            self.center_of_mass.y = (self.center_of_mass.y * self.total_mass + 
                                    particle.position.y * particle.mass) / total_mass_new
        
        self.total_mass = total_mass_new
    
    def subdivide(self):
        x = self.boundary.x
        y = self.boundary.y
        w = self.boundary.w / 2
        h = self.boundary.h / 2
        
        ne = Rectangle(x + w, y - h, w, h)
        self.northeast = QuadTree(ne, self.capacity)
        
        nw = Rectangle(x - w, y - h, w, h)
        self.northwest = QuadTree(nw, self.capacity)
        
        se = Rectangle(x + w, y + h, w, h)
        self.southeast = QuadTree(se, self.capacity)
        
        sw = Rectangle(x - w, y + h, w, h)
        self.southwest = QuadTree(sw, self.capacity)
        
        self.divided = True
        
        # Redistribute particles to children
        for particle in self.particles:
            self.northeast.insert(particle)
            self.northwest.insert(particle)
            self.southeast.insert(particle)
            self.southwest.insert(particle)
        
        # Clear the particles at this node
        self.particles = []
    
    def calculate_force(self, particle, theta):
        if self.total_mass == 0:
            return 0, 0
        
        # If this is an external node (leaf with one particle) and it's the same particle, skip
        if len(self.particles) == 1 and self.particles[0] == particle:
            return 0, 0
        
        # Calculate distance between particle and center of mass
        dx = self.center_of_mass.x - particle.position.x
        dy = self.center_of_mass.y - particle.position.y
        distance = math.sqrt(dx*dx + dy*dy)
        
        # If this is a leaf node or the distance is sufficient for approximation
        if not self.divided or (self.boundary.w * 2) / distance < theta:
            # Avoid division by zero and extreme forces at small distances
            if distance < SOFTENING:
                distance = SOFTENING
                
            # Calculate gravitational force
            f = G * particle.mass * self.total_mass / (distance * distance)
            
            # Resolve force into x and y components
            fx = f * dx / distance
            fy = f * dy / distance
            
            return fx, fy
        
        # Otherwise, recursively calculate forces from children
        fx, fy = 0, 0
        
        if self.northeast:
            fx_ne, fy_ne = self.northeast.calculate_force(particle, theta)
            fx += fx_ne
            fy += fy_ne
        
        if self.northwest:
            fx_nw, fy_nw = self.northwest.calculate_force(particle, theta)
            fx += fx_nw
            fy += fy_nw
        
        if self.southeast:
            fx_se, fy_se = self.southeast.calculate_force(particle, theta)
            fx += fx_se
            fy += fy_se
        
        if self.southwest:
            fx_sw, fy_sw = self.southwest.calculate_force(particle, theta)
            fx += fx_sw
            fy += fy_sw
        
        return fx, fy

def create_deterministic_galaxy(num_particles, center_x, center_y, radius=300, spiral_factor=0.1):
    """Create a deterministic galaxy-like distribution of particles"""
    particles = []
    
    # Create central bulge (30% of particles)
    bulge_count = int(num_particles * 0.3)
    for i in range(bulge_count):
        # Use deterministic distribution for the bulge
        # Distribute in a spiral pattern from center
        distance_factor = i / bulge_count
        r = radius / 5 * distance_factor
        angle = 2 * math.pi * i / bulge_count * 7  # 7 rotations for central bulge
        
        x = center_x + r * math.cos(angle)
        y = center_y + r * math.sin(angle)
        
        # Deterministic mass values
        mass = 50 + (50 * (1 - distance_factor))
        
        particle = Particle(x, y, mass)
        particles.append(particle)
    
    # Create spiral arms (70% of particles)
    spiral_count = num_particles - bulge_count
    arms = 2  # Number of spiral arms
    particles_per_arm = spiral_count // arms
    
    for arm in range(arms):
        base_angle = 2 * math.pi * arm / arms
        
        for i in range(particles_per_arm):
            # Distance from center (linearly distributed from inner to outer radius)
            distance_factor = i / particles_per_arm
            distance = radius * (0.2 + 0.8 * distance_factor)
            
            # Spiral angle based on distance from center
            angle = base_angle + spiral_factor * distance
            
            # Deterministic variation
            variation = 0.2 * math.sin(i * 0.1)
            angle += variation
            
            x = center_x + distance * math.cos(angle)
            y = center_y + distance * math.sin(angle)
            
            # Deterministic mass values
            mass = 1 + 9 * (1 - distance_factor)
            
            particle = Particle(x, y, mass)
            particles.append(particle)
    
    # Remaining particles (if odd number doesn't divide evenly into arms)
    remaining = spiral_count - (particles_per_arm * arms)
    for i in range(remaining):
        angle = 2 * math.pi * i / remaining
        distance = radius * 0.5
        
        x = center_x + distance * math.cos(angle)
        y = center_y + distance * math.sin(angle)
        
        particle = Particle(x, y, 5.0)  # Fixed mass
        particles.append(particle)
    
    # Add deterministic initial orbital velocities
    for p in particles:
        # Vector from center to particle
        dx = p.position.x - center_x
        dy = p.position.y - center_y
        distance = math.sqrt(dx*dx + dy*dy)
        
        if distance > 0:
            # Direction perpendicular to radial direction
            perp_x = -dy / distance
            perp_y = dx / distance
            
            # Orbital velocity (approximating balanced centripetal force)
            orbital_velocity = math.sqrt(0.1 * (distance + 100)) * 0.3
            
            p.velocity.x = perp_x * orbital_velocity
            p.velocity.y = perp_y * orbital_velocity
    
    return particles

def calculate_system_energy(particles):
    """Calculate the total energy of the system (kinetic + potential)"""
    energy = 0.0
    
    # Calculate potential energy
    for i in range(len(particles)):
        for j in range(i + 1, len(particles)):
            p1 = particles[i]
            p2 = particles[j]
            
            dx = p1.position.x - p2.position.x
            dy = p1.position.y - p2.position.y
            distance = math.sqrt(dx*dx + dy*dy)
            
            # Avoid division by zero
            if distance < SOFTENING:
                distance = SOFTENING
            
            # Gravitational potential energy
            energy -= G * p1.mass * p2.mass / distance
    
    # Calculate kinetic energy
    for p in particles:
        v_squared = p.velocity.x * p.velocity.x + p.velocity.y * p.velocity.y
        energy += 0.5 * p.mass * v_squared
    
    return energy

def advance_system(particles, theta, time_step, width, height):
    """Advance the n-body system by one time step using the quadtree"""
    # Create a fresh quadtree
    boundary = Rectangle(width / 2, height / 2, width / 2, height / 2)
    quadtree = QuadTree(boundary)
    
    # Insert all particles into the quadtree
    for particle in particles:
        quadtree.insert(particle)
    
    # Calculate and apply forces to all particles
    for particle in particles:
        fx, fy = quadtree.calculate_force(particle, theta)
        particle.apply_force(fx, fy)
    
    # Update all particles
    for particle in particles:
        particle.update(time_step)

def bench_quadtree_nbody(loops, num_particles, iterations, theta):
    # Initialize simulation space
    width = 1000
    height = 800
    
    # Create deterministic galaxy distribution
    particles = create_deterministic_galaxy(num_particles, width / 2, height / 2)
    
    # Calculate initial energy
    initial_energy = calculate_system_energy(particles)
    
    range_it = range(loops)
    t0 = pyperf.perf_counter()
    
    for _ in range_it:
        # Run simulation for specified iterations
        for _ in range(iterations):
            advance_system(particles, theta, TIME_STEP, width, height)
        
        # Calculate final energy
        final_energy = calculate_system_energy(particles)
    
    return pyperf.perf_counter() - t0

def add_cmdline_args(cmd, args):
    cmd.extend(("--iterations", str(args.iterations)))
    cmd.extend(("--particles", str(args.particles)))
    cmd.extend(("--theta", str(args.theta)))

if __name__ == '__main__':
    runner = pyperf.Runner(add_cmdline_args=add_cmdline_args)
    runner.metadata['description'] = "Quadtree N-body benchmark"
    
    runner.argparser.add_argument("--iterations",
                                  type=int, default=DEFAULT_ITERATIONS,
                                  help="Number of simulation steps per benchmark loop "
                                       "(default: %s)" % DEFAULT_ITERATIONS)
    
    runner.argparser.add_argument("--particles",
                                  type=int, default=DEFAULT_PARTICLES,
                                  help="Number of particles in the simulation "
                                       "(default: %s)" % DEFAULT_PARTICLES)
    
    runner.argparser.add_argument("--theta",
                                  type=float, default=DEFAULT_THETA,
                                  help="Barnes-Hut approximation threshold "
                                       "(default: %s)" % DEFAULT_THETA)
    
    args = runner.parse_args()
    runner.bench_time_func('quadtree_nbody', bench_quadtree_nbody,
                          args.particles, args.iterations, args.theta)



"""
Benchmark for recursive coroutines.

Author: Kumar Aditya
"""

import pyperf


async def fibonacci(n: int) -> int:
    if n <= 1:
        return n
    return await fibonacci(n - 1) + await fibonacci(n - 2)


def bench_coroutines(loops: int) -> float:
    range_it = range(loops)
    t0 = pyperf.perf_counter()
    for _ in range_it:
        coro = fibonacci(25)
        try:
            while True:
                coro.send(None)
        except StopIteration:
            pass
    return pyperf.perf_counter() - t0


if __name__ == "__main__":
    runner = pyperf.Runner()
    runner.metadata['description'] = "Benchmark coroutines"
    runner.bench_time_func('coroutines', bench_coroutines)



"""
A dummy benchmark that uses a local wheel
"""

import pyperf


def bench():
    return 1.0


if __name__ == "__main__":
    runner = pyperf.Runner()
    runner.metadata["description"] = (
        "A dummy benchmark that has a local wheel dependency"
    )

    args = runner.parse_args()
    runner.bench_func("local_wheel", bench)



__all__ = [
    "iter_selections",
    "parse_selection",
]


from . import _benchmark, _utils


def parse_selection(selection, *, op=None):
    # "selection" is one of the following:
    # * a benchmark string
    # * a benchmark name
    # * a benchmark pattern
    # * a tag
    # * a tag pattern
    parsed = _benchmark.parse_benchmark(selection, fail=False)
    spec, metafile = parsed if parsed else (None, None)
    if parsed and spec.version:
        kind = "benchmark"
        spec, metafile = parsed
        if metafile:
            parsed = _benchmark.Benchmark(spec, metafile)
        else:
            parsed = spec
    elif parsed and (spec.origin or metafile):
        raise NotImplementedError(selection)
    else:
        parsed = _utils.parse_tag_pattern(selection)
        if parsed:
            kind = "tag"
        else:
            kind = "name"
            parsed = _utils.parse_name_pattern(selection, fail=True)
            # parsed = _utils.parse_name_pattern(selection, fail=False)
            if not parsed:
                raise ValueError(f"unsupported selection {selection!r}")
    return op or "+", selection, kind, parsed


def iter_selections(manifest, selections, *, unique=True):
    byname = {b.name: b for b in manifest.benchmarks}

    # Compose the expanded include/exclude lists.
    seen = set()
    included = []
    excluded = set()
    for op, _, kind, parsed in selections:
        matches = _match_selection(manifest, kind, parsed, byname)
        if op == "+":
            for bench in matches:
                if bench not in seen or not unique:
                    included.append(bench)
                    seen.add(bench)
        elif op == "-":
            for bench in matches:
                excluded.add(bench)
        else:
            raise NotImplementedError(op)
    if not included:
        included = list(_match_selection(manifest, "tag", "default", byname))

    for bench in included:
        if bench not in excluded:
            yield bench


#######################################
# internal implementation


def _match_selection(manifest, kind, parsed, byname):
    if kind == "benchmark":
        bench = parsed
        # XXX Match bench.metafile too?
        spec = getattr(bench, "spec", bench)
        # For now we only support selection by name.
        # XXX Support selection by version?
        # XXX Support selection by origin?
        if spec.version or spec.origin:
            raise NotImplementedError(spec)
        if spec.name in byname:
            yield bench
        else:
            # No match!  The caller can handle this as they like.
            yield str(bench)
    elif kind == "tag":
        groups = []
        if callable(parsed):
            match_tag = parsed
            for group in manifest.groups:
                if match_tag(group):
                    groups.append(group)
        elif parsed in manifest.groups:
            groups.append(parsed)
        else:
            raise ValueError(f"unsupported selection {parsed!r}")
        for group in groups:
            yield from manifest.resolve_group(group)
    elif kind == "name":
        if callable(parsed):
            match_bench = parsed
            for bench in manifest.benchmarks:
                if match_bench(bench.name):
                    yield bench
        else:
            name = parsed
            if name in byname:
                yield byname[name]
            # We also check the groups, for backward compatibility.
            elif name in manifest.groups:
                yield from _match_selection(manifest, "tag", name, byname)
            else:
                _utils.check_name(name)
                # No match!  The caller can handle this as they like.
                yield name
    else:
        raise NotImplementedError(kind)



"""Simple, brute-force N-Queens solver."""

import pyperf

__author__ = "collinwinter@google.com (Collin Winter)"


# Pure-Python implementation of itertools.permutations().
def permutations(iterable, r=None):
    """permutations(range(3), 2) --> (0,1) (0,2) (1,0) (1,2) (2,0) (2,1)"""
    pool = tuple(iterable)
    n = len(pool)
    if r is None:
        r = n
    indices = list(range(n))
    cycles = list(range(n - r + 1, n + 1))[::-1]
    yield tuple(pool[i] for i in indices[:r])
    while n:
        for i in reversed(range(r)):
            cycles[i] -= 1
            if cycles[i] == 0:
                indices[i:] = indices[i + 1:] + indices[i:i + 1]
                cycles[i] = n - i
            else:
                j = cycles[i]
                indices[i], indices[-j] = indices[-j], indices[i]
                yield tuple(pool[i] for i in indices[:r])
                break
        else:
            return


# From http://code.activestate.com/recipes/576647/
def n_queens(queen_count):
    """N-Queens solver.

    Args:
        queen_count: the number of queens to solve for. This is also the
            board size.

    Yields:
        Solutions to the problem. Each yielded value is looks like
        (3, 8, 2, 1, 4, ..., 6) where each number is the column position for the
        queen, and the index into the tuple indicates the row.
    """
    cols = range(queen_count)
    for vec in permutations(cols):
        if (queen_count == len(set(vec[i] + i for i in cols))
                        == len(set(vec[i] - i for i in cols))):
            yield vec


def bench_n_queens(queen_count):
    list(n_queens(queen_count))


if __name__ == "__main__":
    runner = pyperf.Runner()
    runner.metadata['description'] = "Simple, brute-force N-Queens solver"

    queen_count = 8
    runner.bench_func('nqueens', bench_n_queens, queen_count)



"""
Convert Docutils' documentation from reStructuredText to <format>.
"""

import contextlib
from pathlib import Path

import docutils
from docutils import core
import pyperf

try:
    from docutils.utils.math.math2html import Trace
except ImportError:
    pass
else:
    Trace.show = lambda message, channel: ...  # don't print to console

DOC_ROOT = (Path(__file__).parent / "data" / "docs").resolve()


def build_html(doc_root):
    elapsed = 0
    for file in doc_root.rglob("*.txt"):
        file_contents = file.read_text(encoding="utf-8")
        t0 = pyperf.perf_counter()
        with contextlib.suppress(docutils.ApplicationError):
            core.publish_string(source=file_contents,
                                reader_name="standalone",
                                parser_name="restructuredtext",
                                writer_name="html5",
                                settings_overrides={
                                    "input_encoding": "unicode",
                                    "output_encoding": "unicode",
                                    "report_level": 5,
                                })
        elapsed += pyperf.perf_counter() - t0
    return elapsed


def bench_docutils(loops, doc_root):
    runs_total = 0
    for _ in range(loops):
        runs_total += build_html(doc_root)
    return runs_total


if __name__ == "__main__":
    runner = pyperf.Runner()

    runner.metadata['description'] = "Render documentation with Docutils"
    args = runner.parse_args()

    runner.bench_time_func("docutils", bench_docutils, DOC_ROOT)



import sys

try:
    import pyperformance
except ModuleNotFoundError:
    sys.exit(0)
else:
    print(pyperformance.PKG_ROOT)

# Make sure pyperformance.PKG_ROOT matches expectations.
import os.path

datadir = os.path.dirname(os.path.abspath(__file__))
testsroot = os.path.dirname(datadir)
pkgroot = os.path.dirname(testsroot)
reporoot = os.path.realpath(os.path.dirname(pkgroot))
marker = os.path.join(reporoot, "pyproject.toml")
if not os.path.exists(marker):
    sys.exit(f"ERROR: pyperformance is not an editable install ({reporoot})")
actual = os.path.realpath(os.path.abspath(pyperformance.PKG_ROOT))
if actual != os.path.join(reporoot, "pyperformance"):
    print("ERROR: mismatch on pyperformance repo root:")
    print(f"  actual:   {actual}")
    print(f"  expected: {reporoot}")
    sys.exit(1)
